import os,sys,glob,json,shutil,copy,traceback
from abacustest import constant
from abacustest.lib_prepare.abacus import AbacusStru,ReadInput,WriteInput,WriteKpt
from abacustest.lib_collectdata.collectdata import RESULT
from abacustest.prepare import PrepareAbacus
from ..model import Model
from . import comm
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE

PREPARE_DOC="""
    This model is used to do the finite difference test of force. 
    Should prepare a file named 'info.txt' in the folder of inputs, which contains the label, number of atom, x/y/z component of cell vector, such as:
    '
    C 2 x y z
    H 1 x y 
    O 3 y
    '
    which means do the finite difference test of force 2-nd C on x y z component, and 1-st H on x y component, and 3-rd O on y component.
"""
POST_DOC="""
    the subfolder of the job should be like:
    fdf: the original input
    fdf_<LABEL_NAME>_<LABEL_IDX>_<x/y/z>_<+/->_<inum>: the input with cell vector changed by +/- diff
    there should be a file named "step.txt" in the fdf folder, which contains the step of finite difference test.
"""

class fdforce(Model):
    '''
    Base class for all models
    
    Some specific workflow will be implemented in the subclass.
    Such as EOS, BAND, etc.
    
    The model should have the following methods:
    '''
    @staticmethod
    def model_name():
        '''
        Name of the model, which will be used as the subcommand
        '''
        return "fdforce"
    
    @staticmethod
    def description():
        '''
        Description of the model
        '''
        return "finite difference of force"
    
    @staticmethod
    def add_args(parser):
        '''
        Add arguments for the model
        The arguments can not be command, model, modelcommand 
        '''
        pass
    
    def run(self,params):
        '''
        Parse the parameters and run the model
        '''
        pass
    
    @staticmethod
    def prepare_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand '''
        
        parser.description = PREPARE_DOC
        parser.add_argument('-d', '--step', type=float, default=0.01,help="the step of finite difference (unit: Angstrom), default 0.01")
        parser.add_argument('-n', '--number', type=int, default=5,help="the number of finite difference of one direction, default 5")
        parser.add_argument('-j', '--job',default=["."], action="extend",nargs="*" ,help='the path of abacus inputs, default is current folder')
        
        return parser
    
    def run_prepare(self,params):
        '''
        Parse the parameters and run the prepare process.
        Usually, this step will generate the input files for abacustest submit.
        '''
        jobs = ["."] if len(params.job) == 1 else params.job[1:]
        infof = "info.txt"
        subfolders = PrepareFDForce(infof,jobs,params.step,params.number).run()

        if subfolders:
            setting = {
                "save_path": "results",
                "bohrium_group_name": "FDForce",
                "run_dft": {
                    "example": subfolders,
                    "command": RECOMMAND_COMMAND,
                    "image": RECOMMAND_IMAGE,
                    "bohrium": {
                        "scass_type": RECOMMAND_MACHINE,
                        "job_type": "container",
                        "platform": "ali",
                    },
                },
            }
            comm.dump_setting(setting)
            comm.doc_after_prepare("FDForce", subfolders, ["setting.json"],has_prepare=False)
    
    @staticmethod
    def postprocess_args(parser):
        '''
        Add arguments for the postprocess subcommand
        The arguments can not be command, model, modelcommand'''
        
        parser.description = POST_DOC
        parser.add_argument('-j', '--job',default=["."], action="extend",nargs="*" ,help='the path of abacus inputs, default is current folder.')
        parser.add_argument('-t', '--type',default=0,type=int ,help='the job type. 0:abaus, 1:qe, 2:vasp. Default 0')
        return parser

    def run_postprocess(self,params):
        '''
        Parse the parameters and run the postprocess process'''
        jobs = ["."] if len(params.job) == 1 else params.job[1:]
        PostProcessFDForce(jobs,params.type).run()
    
class PrepareFDForce:
    def __init__(self,infof,jobs,step,number):
        self.infof = infof
        self.jobs = jobs
        self.step = step
        self.number = number
    
    def run(self):
        step = self.step
        num = self.number
        print("step(Angstrom):",self.step)

        subfolders = []
        for job in self.jobs:
            for ijob in glob.glob(job):
                if not os.path.isdir(ijob):
                    continue
                # remove fdf folders
                for ifile in glob.glob(os.path.join(ijob,"fdf*")):
                    if os.path.isdir(ifile):
                        shutil.rmtree(ifile)
                print("prepare inputs in ",ijob)
                infos = self.read_info(ijob)
                if infos:
                    isub = self.preapre_inputs(infos,ijob,step,num=num)
                    if isub:
                        subfolders.extend(isub)
                else:
                    print("ERROR: preapre failed in",ijob)
        # write relative path of subfolders to example.txt
        return subfolders
    
    def read_info(self,path):
        """
        the txt.info should be like:
        C 2 x y z
        H 1 x y
        """
        def print_error(path,line):
            print("ERROR: info.txt format error in ",path)
            print("line:",line)
            print(PREPARE_DOC)
        if os.path.isfile(os.path.join(path,self.infof)):
            with open(os.path.join(path,self.infof)) as f:
                lines = f.readlines()
                info = []
                for line in lines:
                    if line.strip() == "":
                        continue
                    sline = line.strip().split()
                    if len(sline) < 3 or not sline[1].isdigit():
                        print_error(path,line)
                        continue
                    xyz = [i.lower() for i in sline[2:]]
                    # if xyz has other character, then error
                    has_error = False
                    for i in xyz:
                        if i not in ["x","y","z"]:
                            print_error(path,line)
                            has_error = True
                            break
                    if int(sline[1]) <= 0:
                        print("ERROR: atom index should be start from 1")
                        print_error(path,line)
                        has_error = True
                    if not has_error:
                        info.append([sline[0],int(sline[1]),xyz])
                return info
        else:
            print("ERROR: info.txt not found in ",path)
            return None
        
    def prepare_one_stru_abacus(self,istru,final_path,otherfiles,input_param):
        os.makedirs(final_path,exist_ok=True)
        istru.write(os.path.join(final_path,"STRU"))
        # copy other files to final_path
        for file in otherfiles:
            os.symlink(file,os.path.join(final_path,os.path.basename(file)))

        # write input
        PrepareAbacus.WriteInput(input_param,os.path.join(final_path,"INPUT"))

    def preapre_inputs(self,infos, init_path, step = 0.001, num=3) :
        subfolders = []
        input_param = {"calculation": "scf","cal_force":1}
        otherfiles = []
        for ifile in os.listdir(init_path):
            if ifile == "INPUT":
                input_param = PrepareAbacus.ReadInput(os.path.join(init_path,"INPUT"))
                input_param["calculation"] = "scf"
                input_param["cal_force"] = 1
                continue
            elif ifile != "STRU" and not ifile.startswith(".") and os.path.isfile(os.path.join(init_path,ifile)):
                otherfiles.append(os.path.abspath(os.path.join(init_path,ifile)))

        stru = AbacusStru.ReadStru(os.path.join(init_path,"STRU"))
        if not stru:
            print(f"ERROR: read STRU failed in {init_path}")
            return None

        label = stru.get_label()
        atom_num = [label.count(i) for i in label] # the number of atoms for each element
        coord = stru.get_coord(direct=False,bohr=False)
        cell = stru.get_cell(bohr=True)  # need use cell as the input of AbacusStru
        new_stru = copy.deepcopy(stru)
        new_stru.set_coord(coord,bohr=False,direct=False)

        # write original input
        self.prepare_one_stru_abacus(new_stru,os.path.join(init_path,"fdf"),otherfiles,input_param)
        with open(os.path.join(init_path,"fdf","step.txt"),"w") as f: f.write(str(step))
        subfolders.append(os.path.join(init_path,"fdf"))

        xyz_idx = {"x":0,"y":1,"z":2}
        # write other inputs
        for info in infos:
            label_name, label_idx, xyz = info
            if label_name not in label:
                print(f"ERROR: label {label_name} not in STRU")
                continue
            if label_idx > atom_num[label.index(label_name)]: 
                print(f"ERROR: label {label_name} has only {atom_num[label.index(label_name)]} atoms")
                continue
            
            atom_idx = 0
            for i in range(len(label)):
                if label[i] == label_name:
                    atom_idx += 1
                    if atom_idx == label_idx:
                        break
            atom_idx = i

            for ixyz in xyz:
                for inum in range(num):
                    subfolder_name1 = f"fdf_{label_name}_{label_idx}_{ixyz}_+_{inum+1}"
                    subfolder_name2 = f"fdf_{label_name}_{label_idx}_{ixyz}_-_{inum+1}"
                    subfolders.append(os.path.join(init_path,subfolder_name1))
                    subfolders.append(os.path.join(init_path,subfolder_name2))

                    # write stru
                    new_coord = copy.deepcopy(coord)

                    new_coord[atom_idx][xyz_idx[ixyz]] += step * (inum+1)
                    new_stru = copy.deepcopy(stru)
                    new_stru.set_coord(new_coord,bohr=False,direct=False)
                    self.prepare_one_stru_abacus(new_stru,
                            os.path.join(init_path,subfolder_name1),
                            otherfiles,
                            input_param)

                    new_coord = copy.deepcopy(coord)
                    new_coord[atom_idx][xyz_idx[ixyz]] -= step * (inum+1)
                    new_stru = copy.deepcopy(stru)
                    new_stru.set_coord(new_coord,bohr=False,direct=False)
                    self.prepare_one_stru_abacus(new_stru,
                            os.path.join(init_path,subfolder_name2),
                            otherfiles,
                            input_param)
        return subfolders
    
    
class PostProcessFDForce:
    def __init__(self,jobs,job_type=0):
        self.jobs = jobs
        
        job_type_dict = {0:"abacus",1:"qe",2:"vasp"}
        if job_type not in job_type_dict:
            print(f"ERROR: job type {job_type} not supported")
            job_type = 0
        self.job_type_str = job_type_dict.get(job_type,"abacus")
        
        pass
    
    def run(self):
        allresults,cases = self.get_results()
        json.dump(allresults,open("metrics.json","w"),indent=4)
        
        plot_results = self.gen_plot_results(allresults,cases)
        allpngs = self.plot_force(plot_results)
        
        if allpngs:
            json.dump({i[:-4]: {"type": "image", "file": i} for i in allpngs},open("supermetrics.json","w"),indent=4) 
    
    def get_results(self):
        """
        return allresults, cases
        allresults is a dict with key is all fdf jobs and value is a dict of key metrics
        cases is a dict with key is the case name and value is a dict of key metrics
        """
        allresults = {}
        cases = {}
        for ijob in self.jobs:
            for iijob in glob.glob(ijob):
                if not os.path.isdir(iijob):
                    continue
                if not os.path.isdir(os.path.join(iijob,"fdf")):
                    print(f"ERROR: {iijob} does not have fdf folder")
                    continue
                if not os.path.isfile(os.path.join(iijob,"fdf","step.txt")):
                    print(f"ERROR: {iijob} does not have step.txt")
                    continue
                
                step = float(open(os.path.join(iijob,"fdf","step.txt")).readline()) # the unit from step.txt is Angstrom
                print(f"get results in {iijob}")
                for job in glob.glob(os.path.join(iijob,"fdf*")):
                    job_name = os.path.basename(job)
                    job_name_list = job_name.split("_")
                    # job_name should be like fdf or fdf_Fe_x_1_+_1
                    if os.path.isdir(job) and (job_name == "fdf" or len(job_name_list) == 6):
                        r = RESULT(path=job,fmt=self.job_type_str)
                        allresults[job] = {
                            "converge": r["converge"],
                            "energy": r["energy"],
                            "force": r["force"],
                            "drho_last": r["drho_last"],
                            "denergy_last": r["denergy_last"],
                            "atomlabel_list": r["atomlabel_list"],
                            "total_time": r["total_time"]       
                        }
                        if len(job_name_list) == 6:
                            icasename = os.path.join(iijob,"_".join(job_name_list[1:4]))
                            if icasename not in cases:
                                cases[icasename] = {"step": step, "number": int(job_name_list[-1]),"label": job_name_list[1], "idx": int(job_name_list[2]), "xyz": job_name_list[3],
                                                    "label_idx":None}
                            if r["atomlabel_list"] != None and cases[icasename]["label_idx"] == None:
                                label_idx = None
                                idx = 0
                                for i in range(len(r["atomlabel_list"])):
                                    if r["atomlabel_list"][i] == job_name_list[1]:
                                        idx += 1
                                    if idx == int(job_name_list[2]):
                                        label_idx = i
                                        break
                                cases[icasename]["label_idx"] = label_idx
                                
                            if int(job_name_list[-1]) > cases[icasename]["number"]:
                                cases[icasename]["number"] = int(job_name_list[-1])
        return allresults,cases        

    def gen_fdf_name(self,example_name,case_name,step_num):
        # case name should be like: Fe_1_x
        if step_num == 0:
            return os.path.join(example_name,"fdf")
        elif step_num > 0:
            return os.path.join(example_name,f"fdf_{case_name}_+_{step_num}")
        else:
            return os.path.join(example_name,f"fdf_{case_name}_-_{-step_num}")

    def gen_plot_results(self,allresults,cases):
        if not allresults:
            return {}
        plot_results = {}  # results for FD at different STRU with same step
        
        for case in cases:
            example_name = os.path.dirname(case)
            case_name = os.path.basename(case)
            input_r = allresults.get(os.path.join(example_name,"fdf"),{})  # the results of input structure
            step = cases[case]["step"]
            maxnumber = cases[case]["number"]
            label_idx = cases[case]["label_idx"]
            xyz_idx = {"x":0, "y":1, "z":2}[cases[case]["xyz"]]
            if label_idx == None:
                print(f"ERROR: {case} is failed")
                continue
            force_idx = label_idx * 3 + xyz_idx
            
            plot_results[case] = {"r1":{"pos": [], "energy": [], "force": [], "fd_force": []},
                                  "r2":{"step": [], "fd_force": [], "force": None if input_r.get("force") == None else input_r.get("force")[force_idx]}}
            
            # plot_results1
            for i in range(-maxnumber+1,maxnumber):
                fdf_name = self.gen_fdf_name(example_name,case_name,i)
                fdf_name_p = self.gen_fdf_name(example_name,case_name,i+1)
                fdf_name_m = self.gen_fdf_name(example_name,case_name,i-1)
                force = allresults.get(fdf_name,{}).get("force")
                energy_p = allresults.get(fdf_name_p,{}).get("energy")
                energy_m = allresults.get(fdf_name_m,{}).get("energy")
                
                plot_results[case]["r1"]["pos"].append(i*step)
                plot_results[case]["r1"]["energy"].append(allresults.get(fdf_name,{}).get("energy"))
                plot_results[case]["r1"]["force"].append(None if force == None else force[force_idx])
                if energy_p != None and energy_m != None:
                    plot_results[case]["r1"]["fd_force"].append((energy_m - energy_p) / (2 * step))
                else:
                    plot_results[case]["r1"]["fd_force"].append(None)
            
            # plot_results2
            for i in range(1,maxnumber+1):
                fdf_name_p = self.gen_fdf_name(example_name,case_name,i)
                fdf_name_m = self.gen_fdf_name(example_name,case_name,-i)
                energy_p = allresults.get(fdf_name_p,{}).get("energy")
                energy_m = allresults.get(fdf_name_m,{}).get("energy")
                plot_results[case]["r2"]["step"].append(i*step*2)
                if energy_p != None and energy_m != None:
                    plot_results[case]["r2"]["fd_force"].append((energy_m - energy_p) / (2 * step * i))
                else:
                    plot_results[case]["r2"]["fd_force"].append(None)
            
            json.dump(plot_results[case],open(os.path.join(example_name,f"{case_name}.json"),"w"),indent=4)
                
        return plot_results

    def cal_rmsd(self,list1,list2):
        l1,l2 = comm.clean_none_list(list1,list2)
        if len(l1) == 0:
            return None
        return (sum([(i-j)**2 for i,j in zip(l1,l2)]) / len(l1))**0.5
    
    def plot_force(self, results):
        import matplotlib.pyplot as plt
        from abacustest.lib_model.comm_plot import set_font
        set_font("Times New Roman",12)
        
        allpngs = []
        for case in results:
            distance = results[case]["r1"]["pos"]
            energy = results[case]["r1"]["energy"]
            force = results[case]["r1"]["force"]
            fd_force = results[case]["r1"]["fd_force"]
            
            d,e,f,fd = comm.clean_none_list(distance,energy,force,fd_force)
            x, y = comm.clean_none_list(results[case]["r2"]["step"],results[case]["r2"]["fd_force"])
            if len(d) == 0 or len(x) == 0:
                continue
            
            fig,ax = plt.subplots(1,2,figsize=(12,5))
            ax1 = ax[0]
            #fig, ax1 = plt.subplots()
            ax2 = ax1.twinx()
            ax1.plot(d,e,"b+-",label="Energy(eV)", markersize=10)
            ax2.plot(d,f,"ro--",label="Analytic(eV/$\mathrm{\AA}$)", markersize=10)
            ax2.plot(d,fd,"m*--",label="Finite Difference(eV/$\mathrm{\AA}$)", markersize=10)
            
            ymax1 = max(e)
            ymin1 = min(e)
            ymax2 = max(max(f),max(fd))
            ymin2 = min(min(f),min(fd))
            ax1.set_xlabel("Atomic position ($\mathrm{\AA}$)" +  f"(FD step size: {x[0]:.3f}" + " $\mathrm{\AA}$)")
            ax1.set_ylabel("Energy (eV)",color="b")
            ax2.set_ylabel("Force (eV/$\mathrm{\AA}$)",color="r")
            ax1.spines['left'].set_color('blue')
            ax1.tick_params(axis='y', colors='blue')
            ax2.spines['right'].set_color('red')
            ax2.tick_params(axis='y', colors='red')
            ax1.set_ylim(ymin1-(ymax1-ymin1)*0.1,ymax1+(ymax1-ymin1)*0.3)
            ax2.set_ylim(ymin2-(ymax2-ymin2)*0.1,ymax2+(ymax2-ymin2)*0.3)
            rmsd = self.cal_rmsd(f,fd)
            ax1.set_title(case + f" (Deviation RMSD={rmsd:.2e} " + "eV/$\mathrm{\AA}$)")
            print(case + f" (Deviation RMSD={rmsd:.2e} eV/Angstrom)")
            ax1.legend(loc="upper left",frameon=False)
            ax2.legend(loc="upper right",frameon=False)
            ax2.grid(True)

            ax3 = ax[1]
            # plot force0-FD vs distance0
            y_ref = results[case]["r2"]["force"]
            ymin = min(y_ref, min(y))
            ymax = max(y_ref, max(y))
            # sort x,y by x
            ax3.plot(x,y,"m*--",label="Finite Difference(eV/$\mathrm{\AA}$)", markersize=10)
            if y_ref != None:
                ax3.axhline(y_ref,ls="--",color="red",label="Analytic(eV/$\mathrm{\AA}$)")
            ax3.set_xlabel("FD step size ($\mathrm{\AA}$)")
            ax3.set_ylabel("Force (eV/$\mathrm{\AA}$)")
            ax3.set_title(case + f" (FD at initial position VS step size)")
            ax3.legend(loc="upper right",frameon=False)
            ax3.grid(True)
            ax3.set_ylim(ymin-(ymax-ymin)*0.1,ymax+(ymax-ymin)*0.3)

            plt.subplots_adjust(right=0.85) 
            png = case + ".png"
            plt.tight_layout()
            plt.savefig(png,dpi=300)
            plt.close()
            allpngs.append(png)
        return allpngs


import traceback
try:
    from abacustest.lib_prepare.abacus import ReadInput, WriteInput, AbacusStru, ReadKpt, WriteKpt
    from abacustest.lib_collectdata.collectdata import RESULT
except ImportError:
    traceback.print_exc()
    print("Import Error in abacustest/__init__.py")


from ..model import Model
import json, os
from abacustest.lib_prepare.abacus import WriteKpt, WriteInput, ReadInput, AbacusStru
from abacustest.lib_prepare.comm import kspacing2kpt
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE

from typing import List, Dict, Any
from typing import Union
import warnings
import shutil, glob, copy, traceback
from abacustest.lib_collectdata.collectdata import RESULT

from abacustest.outresult import pandas_out


class ElasticModel(Model):
    @staticmethod
    def model_name(): # type: ignore
        '''
        Name of the model, which will be used as the subcommand
        '''
        return "elastic"
    
    @staticmethod
    def description(): # type: ignore
        '''
        Description of the model
        '''
        return "Prepare and postprocess the elastic"
    
    @staticmethod
    def prepare_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand 
        '''
        parser.description = "Prepare the inputs for the elastic calculation."
        parser.add_argument('-j', '--job', default=[], action="extend", nargs="*", help='the paths of ABACUS jobs, should contain INPUT, STRU, or KPT, and pseudopotential and orbital files')
        parser.add_argument("--norm", default=0.01, type=float, help="The maximum strain for normal mode, default is 0.01")
        parser.add_argument("--shear", default=0.01, type=float, help="The maximum strain for shear mode, default is 0.01")
        parser.add_argument("--norelax", action="store_true", help="Whether to not do atomic relaxation for each deformed structure, default is doing.")
        parser.add_argument("--image", type=str, default=RECOMMAND_IMAGE, help="The image to use for the Bohrium job, default is %s" % RECOMMAND_IMAGE)
        parser.add_argument("--machine", type=str, default=RECOMMAND_MACHINE, help="The machine to use for the Bohrium job, default is 'c32_m64_cpu'.")
        parser.add_argument("--abacus_command", type=str, default=RECOMMAND_COMMAND, help=f"The command to run the Abacus job, default is '{RECOMMAND_COMMAND}'.")
        return parser
    
    
    def run_prepare(self,params):
        '''
        Run the model with the given parameters
        '''
        if not params.job:
            raise ValueError("No job specified, please use -j or --job to specify the job paths.")
        
        paths = PrepElastic(params.job, norm=params.norm, shear=params.shear, norelax=params.norelax).run()
        
        setting = {
            "save_path": "results",
            "run_dft": []
        }
        
        for path in paths:
            setting["run_dft"].append({
                "ifrun": True,
                "example": path,
                "command": params.abacus_command,
                "image": params.image,
                "bohrium": {
                    "scass_type": params.machine,
                    "job_type": "container",
                    "platform": "ali"
                }
            })
        
        setting_file = "setting.json"
        json.dump(setting, open(setting_file, "w"), indent=4)
        print("\nThe inputs are generated in", ", ".join(params.job))
        print(f"You can modify '{setting_file}', and execute below command to run the abacustest to submit all jobs to bohrium:\n\tabacustest submit -p setting.json\n")
        print(f"After finishing the calculations, you can enter the results directory, \nand run below command below to postprocess the elastic results:\n\tabacustest model elastic post -j {' '.join(params.job)}\n")
        
    
    def postprocess_args(parser):
        '''
        Add arguments for the postprocess subcommand
        The arguments can not be command, model, modelcommand 
        '''
        parser.description = "Postprocess the elastic calculation results."
        parser.add_argument('-j', '--job', default=[], action="extend", nargs="*", help='the paths of the job directories, should contain the results of deformed structures generated by the prepare step.')
        return parser   

    def run_postprocess(self,params):
        '''
        Parse the parameters and run the postprocess process'''
        if not params.job:
            raise ValueError("No job specified, please use -j or --job to specify the job paths.")
        
        metrics, results = PostprocessElastic(params.job).run()
        json.dump(metrics, open("metrics.json", "w"), indent=4)
        json.dump(results, open("metrics_elastic.json", "w"), indent=4)
        pandas_out(results, "metrics_elastic.csv",print_list_seperate=True)
        print("\nThe postprocess is done. The metrics are saved in 'metrics.json', and the elastic results are saved in 'metrics_elastic.json'.")
        


class PrepElastic():
    """Prepare the elastic calculation.
    
    Args:
        job (str or list of str): the path of the job directory or a list of job directories.
        norm (float): the maximum strain for normal mode, default is 0.01
        shear (float): the maximum strain for shear mode, default is 0.01
        relax (bool): whether to do atomic relaxation for each deformed structure, default is False
    
    run() function can generate the deformed structures and write it into the job directories, and also link the original files to the deformed directories.
    It will return a list of paths of the deformed directories.
    The deformed directories will be named as "deformed_00", "deformed_01", ..., "deformed_N" where N is the number of deformed structures.
    """
    def __init__(self, job: Union[str, List[str]], 
                 norm: float = 0.01, 
                 shear: float = 0.01,
                 norelax: bool = False):
        self.job = job if isinstance(job, list) else [job]
        self.norm_strain = norm
        self.shear_strain = shear
        self.norelax = norelax
    
    def run(self):
        paths = []
        for job in self.job:
            if not os.path.isdir(job):
                warnings.warn(f"Job {job} is not a directory, skip it.")
                continue
            print(f"Preparing elastic calculation for job: {job}")
            paths.append(self.prepare_one_job(job))
        
        return paths 
            
    
    def prepare_one_job(self, job: str):
        '''
        Prepare the elastic calculation for one job
        '''
        # clean the deformed paths
        shutil.rmtree(os.path.join(job,"org"), ignore_errors=True)
        for i in glob.glob(os.path.join(job, "deformed_*")):
            shutil.rmtree(i, ignore_errors=True)
        
        stru = AbacusStru.ReadStru(os.path.join(job, "STRU"))
        
        # modify INPUT file
        input_param = ReadInput(os.path.join(job, "INPUT"))
        if self.norelax:
            input_param["calculation"] = "scf"
        else:
            input_param["calculation"] = "relax"
        input_param["cal_stress"] = 1
            
        if "kspacing" in input_param:
            kpt = kspacing2kpt(float(input_param["kspacing"]), stru.get_cell(bohr=True))
            warnings.warn(f"Transform kspacing {input_param['kspacing']} to kpt {kpt}.")
            WriteKpt(kpt, os.path.join(job, "KPT"))
            del input_param["kspacing"]

        org_files = os.listdir(job)
        for ifile in ["INPUT", "STRU"]:
            if ifile in org_files:
                org_files.remove(ifile)   
        
        # generate the deformed structures
        deformed_strus, strains = self._gen_deformed_structure(stru)

        org_path = os.path.join(job, "org")
        self._gen_subjob(job, org_path, org_files, input_param, stru)
        
        job_paths = [org_path]
        deformed_paths = []
        for i, def_stru in enumerate(deformed_strus):
            deformed_path = f"deformed_{i:02d}"
            ipath = os.path.join(job, deformed_path)
            self._gen_subjob(job, ipath, org_files, input_param, def_stru)
            job_paths.append(ipath)
            deformed_paths.append(deformed_path)
        
        json.dump({"strains": [s.as_dict() for s in strains],
                   "deformed_paths": deformed_paths}, 
                    open(os.path.join(org_path, "elastic.json"), "w"),
                          indent=4)

        return job_paths    

    def _gen_deformed_structure(self, stru: AbacusStru):
        from pymatgen.analysis.elasticity.strain import DeformedStructureSet
        from pymatgen.analysis.elasticity.elastic import Strain
        
        structure = stru.to_pymatgen()
        
        norm_strains = [-self.norm_strain, -0.5*self.norm_strain, +0.5*self.norm_strain, +self.norm_strain]
        shear_strains = [-self.shear_strain, -0.5*self.shear_strain, +0.5*self.shear_strain, +self.shear_strain]
        
        
        deformed_strus = DeformedStructureSet(structure,
                                    symmetry=False,
                                    norm_strains=norm_strains,
                                    shear_strains=shear_strains)
        
        strus = []
        for defor in deformed_strus:
            new_stru = copy.deepcopy(stru)
            new_stru.set_cell(defor.lattice.matrix, bohr=False)
            new_stru.set_coord(defor.frac_coords, bohr=False, direct=True)
            strus.append(new_stru)
        
        strains = [Strain.from_deformation(d) for d in deformed_strus.deformations]  
        
        return strus, strains
    
    def _gen_subjob(self, 
                    job: str,
                    ipath: str, 
                    org_files: List[str],
                             input_param: Dict[str, Any],
                             new_stru: AbacusStru):
        """Generate a subjob for the elastic calculation.
        
        Args:
            job (str): the path of the job directory.
            ipath (str): the path of the subjob directory.
            org_files (List[str]): the list of original files to link.
            input_param (Dict[str, Any]): the input parameters for the subjob.
            new_stru (AbacusStru): the new structure for the subjob.
        """
        os.makedirs(ipath, exist_ok=True)
        new_stru.write(os.path.join(ipath, "STRU"))
        WriteInput(input_param, os.path.join(ipath, "INPUT"))
        
        # link the original files
        for ifile in org_files:
            os.symlink(os.path.abspath(os.path.join(job, ifile)), os.path.join(ipath, ifile))
        
        

class PostprocessElastic():
    """Postprocess the elastic calculation results.
    
    Args:
        job (str or list of str): the path of the job directory or a list of job directories. 
            Each job should contain directories "org" and "deformed_00", "deformed_01", ..., "deformed_N" where N is the number of deformed structures.
            "org" is the results of the original structure, and a file elastic.json should be present in the "org" directory, which contains the strain values.

    """
    def __init__(self, job: Union[str, List[str]]):
        self.job = job if isinstance(job, list) else [job]
    
    def run(self):
        results = {}
        metrics = {}
        for job in self.job:
            if not os.path.isdir(job):
                warnings.warn(f"Job {job} is not a directory, skip it.")
                continue
            print(f"Postprocessing elastic calculation for job: {job}")
            try:
                m, r = self.postprocess_one_job(job)
            except Exception as e:
                traceback.print_exc()
                warnings.warn(f"Failed to postprocess job {job}: {e}")
                continue
            
            metrics.update(m)
            results[job] = r
        
        return metrics, results
    
    def postprocess_one_job(self, job: str):
        '''
        Postprocess the elastic calculation for one job
        '''
        # check files
        elastic_file = os.path.join(job, "org", "elastic.json")
        original_stru_file = os.path.join(job, "org", "STRU")
        if not os.path.isfile(elastic_file) or not os.path.isfile(original_stru_file):
            raise ValueError(f"Elastic file {elastic_file} or original structure file {original_stru_file} not found in {job}.")
        
        sub_dirs = [os.path.join(job, "org")] + [os.path.join(job, f"deformed_{i:02d}") for i in range(24)]
        
        metrics = self.get_metrics(sub_dirs)
        strain_param = json.load(open(elastic_file))
        strains = strain_param["strains"]
        deformed_paths = strain_param["deformed_paths"]
        
        input_stress = self.abacus_stress_to_pymatgen_stress(metrics[os.path.join(job, "org")]["stress"])
        deformed_stress = [self.abacus_stress_to_pymatgen_stress(metrics[os.path.join(job, dp)]["stress"]) for dp in deformed_paths]
        elastic_results = self.cal_elastic(strains, deformed_stress, input_stress)
        
        return metrics, elastic_results
        
    def abacus_stress_to_pymatgen_stress(self, stress: List[float]):
        """
        Transform calculated stress (units in kBar) collected by abacustest
        to Pymatgen format (units in GPa)
        """
        from pymatgen.analysis.elasticity.stress import Stress
        import numpy as np
        
        return Stress(-0.1 * np.array([stress[0:3],
                                   stress[3:6],
                                   stress[6: ]])) # 1 kBar = 0.1 GPa    
        
    def cal_elastic(self, strain, deformed_stru_stresses, input_stru_stress):
        from pymatgen.analysis.elasticity.elastic import ElasticTensor
        from pymatgen.analysis.elasticity.elastic import Strain
        strain = [Strain.from_dict(s) for s in strain]
        
        result = ElasticTensor.from_independent_strains(strain,
                                                    deformed_stru_stresses,
                                                    eq_stress=input_stru_stress,
                                                    vasp=False)
    
        elastic_tensor = result.voigt.tolist()
        bv, gv = result.k_voigt, result.g_voigt
        ev = 9 * bv * gv / (3 * bv + gv)
        uV = (3 * bv - 2 * gv) / (6 * bv + 2 * gv)

        return {
            "elastic_tensor": elastic_tensor,
            "bulk_modulus": float(bv),
            "shear_modulus": float(gv),
            "young_modulus": float(ev),
            "poisson_ratio": float(uV)
        }    
            
    def get_metrics(self, jobs):
        metrics = {}
        for sub_dir in jobs:
            if not os.path.isdir(sub_dir):
                raise ValueError(f"Subdirectory {sub_dir}.")
            
            r = RESULT(path=sub_dir, fmt="abacus")
            metrics[sub_dir] = {i: r[i] for i in ["normal_end", "converge", "scf_steps", "relax_steps","relax_converge","denergy_last", "drho_last", "energy", "stress"]}

        return metrics      
        
        
        
            
            



import os,sys,glob,re
from ..resultAbacus import ResultAbacus

class BdaAbacus(ResultAbacus):
    
    @ResultAbacus.register(bda_mag_moment="mag_moment of some metal element",
                           bda_bond_length="bond_length of some metal element")
    def GetBDAinfo(self):
        from pymatgen.core.structure import Structure
        from pymatgen.io.vasp.inputs import Poscar
        structure = Structure(lattice=self['cell'],
                              species=self['element_list'],
                              coords=self["coordinate"],
                              coords_are_cartesian=True)
        mag = tuple([{"tot":i} for i in self['atom_mag']])
        from ..comm_funcs.bda import BasicProperty
        output = BasicProperty(mag,None,None,Poscar(structure))
        mag = output.magnetic_moment('TM')
        bond = output.bond_length()
        if mag:
            self["bda_mag_moment"] = mag
        else:
            self["bda_mag_moment"] = None
        if bond:
            self["bda_bond_length"] = bond
        else: 
            self["bda_bond_length"] = None

        
        


from ..model import Model
from . import comm
import argparse,json, os
from abacustest.lib_prepare.abacus import WriteKpt, WriteInput, gen_stru, ReadInput, AbacusStru
from pathlib import Path
from abacustest.lib_model.model_012_band import PrepBand
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE

import warnings

from typing import List, Dict, Union, Optional, Tuple, Literal


JOB_TYPES = {"scf": {"calculation": "scf", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                    "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                    "mixing_beta": 0.8,  "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                    "#cal_force": 1, "#cal_stress": 1,
                    "kspacing": "0.14 # unit in 1/bohr"}, 
             "relax": {"calculation": "relax", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                       "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                       "mixing_beta": 0.8, "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                       "cal_force": 1, "#cal_stress": 1,"kspacing": "0.14 # unit in 1/bohr",
                       "relax_method": "cg # or bfgs bfgs_trad cg_bfgs sd fire",
                       "relax_nmax": 60, "force_thr_ev": "0.01  # unit in eV/A", "#stress_thr": "0.5 # unit in kbar",
                       }, 
             "cell-relax":{"calculation": "cell-relax", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                       "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                       "mixing_beta": 0.8, "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                       "cal_force": 1, "cal_stress": 1, "kspacing": "0.14 # unit in 1/bohr",
                       "relax_method": "cg # or bfgs, bfgs_trad, cg_bfgs, sd, fire",
                       "relax_nmax": 60, "force_thr_ev": "0.01  # unit in eV/A", "stress_thr": "0.5 # unit in kbar",
                       "fixed_axes": "None # or volume, shape, a, b, c, ab, ac, bc; only valid for cell-relax calculation to fix some axes",
                       }, 
             "md":{"calculation": "md", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                       "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                       "mixing_beta": 0.8, "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                       "#cal_force": 1, "#cal_stress": 1,
                       "kspacing": "0.14 # unit in 1/bohr", 
                       "md_type": "nvt  # or npt, nve, langevin, fire, msst",
                       "md_nstep": "10  # number of steps",
                       "md_dt": "1.0  # unit in fs", 
                       "md_tfirst": "100  # unit in K",
                       "md_tlast": "100  # unit in K",}, 
             "band":{"calculation": "scf", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                    "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                    "mixing_beta": 0.8,  "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                    "#cal_force": 1, "#cal_stress": 1,
                    "kspacing": "0.14 # unit in 1/bohr"}
        }

LCAO_PARAM = {
    "basis_type": "lcao",
    "ks_solver": "genelpa",
    "#gamma_only": 0,
    "ecutwfc": 100,
    "scf_thr": 1e-7,
}

ELEMENT_DFTU_D = ["Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn",
                  "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd",
                  "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "La", "Ac", "Th"]
ELEMENT_DFTU_F = ["Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu",
                  "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr"]

MAG_ELEMENT = ELEMENT_DFTU_D + ELEMENT_DFTU_F


class InputsModel(Model):
    @staticmethod
    def model_name(): # type: ignore
        '''
        Name of the model, which will be used as the subcommand
        '''
        return "inputs"
    
    @staticmethod
    def description(): # type: ignore
        '''
        Description of the model
        '''
        return "Prepare the ABACUS inputs of specified model"
    
    @staticmethod
    def add_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand '''
        
        parser.description = "Prepare the ABACUS inputs file."
        parser.add_argument('-f', '--file',default=[], action="extend",nargs="*" ,help='the structure files')
        parser.add_argument("--ftype",default="cif",type=str,help="the structure type, should be cif or dpdata supportted type like: poscar, abacus/stru, ..", )
        parser.add_argument("--jtype", default="scf", type=str, help=f"the job type, should be one of {list(JOB_TYPES.keys())}")
        parser.add_argument("--pp",default=None,type=str,help="the path of pseudopotential library, or read from enviroment variable ABACUS_PP_PATH")
        parser.add_argument("--orb",default=None,type=str,help="the path of orbital library, or read from enviroment variable ABACUS_ORB_PATH")
        parser.add_argument("--input",default=None,type=str,help="the template of input file, if not specified, the default input will be generated")
        parser.add_argument("--kpt", default=None, type=int, nargs="*", help="the kpoint setting, should be one or three integers")
        parser.add_argument("--lcao", action="store_true", help="whether to use lcao basis, default is pw basis")
        parser.add_argument("--nspin", default=1, type=int, choices=[1, 2, 4], help="the number of spins, can be 1 (no spin), 2 (spin polarized), or 4 (non-collinear spin). Default is 1.")
        parser.add_argument("--soc", action="store_true", help="whether to use spin-orbit coupling, if True, nspin should be 4.")
        parser.add_argument("--dftu", action="store_true", help="whether to use DFT+U, default is False.")
        parser.add_argument("--dftu_param", default=None, nargs="+", help="the DFT+U parameters, should be element symbol and U value pairs like 'Fe 4 Ti 1'. If dftu is set, but dftu_param is not set, the default U values will be used: 4 eV for d orbital elements and 6 eV for f orbital elements.")
        parser.add_argument("--init_mag", default=None, nargs="+", help="the initial magnetic moment for magnetic elements, should be element symbol and magnetic moment pairs like 'Fe 4 Ti 1'.")
        parser.add_argument("--afm", action="store_true", help="whether to use antiferromagnetic calculation, default is False. Only valid when init_mag is set.")
        parser.add_argument("--copy_pp_orb", action="store_true", help="whether to copy the pseudopotential and orbital files to each job directory or link them. Default is False, which means linking the files.")
        parser.add_argument("--download-pporb", nargs="?", type=str,default=None,const="apns-v1",choices=["apns-v1"], help="Download the recommended pseudopotential and orbital files from AISquare.")
        return parser
    
    def parse_dftu_param(self, values):
        """
        Parse the values from the command line arguments.
        """
        if len(values) % 2 != 0:
            raise ValueError("The DFT+U parameters should be in pairs of element symbol and U value.")
        dftu_param = {}
        for i in range(0, len(values), 2):
            element = values[i]
            u_value = float(values[i + 1])
            if element not in dftu_param:
                dftu_param[element] = u_value
            else:
                warnings.warn(f"Element {element} already has a DFT+U value, overwriting it with {u_value}.")
        return dftu_param
     
    def parse_init_mag(self, values):
        """
        Parse the initial magnetic moment values from the command line arguments.
        """
        if len(values) % 2 != 0:
            raise ValueError("The initial magnetic moment parameters should be in pairs of element symbol and magnetic moment.")
        init_mag = {}
        for i in range(0, len(values), 2):
            element = values[i]
            mag_value = float(values[i + 1])
            if element not in init_mag:
                init_mag[element] = mag_value
            else:
                warnings.warn(f"Element {element} already has an initial magnetic moment, overwriting it with {mag_value}.")
        return init_mag   
    
    def download_pporb(self, version):
        """
        Download the recommended pseudopotential and orbital files from AISquare.
        """
        if version is None:
            return
        
        apns_v1 = "https://store.aissquare.com/datasets/af21b5d9-19e6-462f-ada1-532f47f165f2/ABACUS-APNS-PPORBs-v1.zip"
        
        if version == "apns-v1":
            # download the apns-v1 pseudopotential and orbital files
            import urllib.request
            import zipfile
            
            zip_file = "ABACUS-APNS-PPORBs-v1.zip"
            print(f"Downloading the recommended pseudopotential and orbital files from {apns_v1} ...")
            urllib.request.urlretrieve(apns_v1, zip_file)
            print(f"Unzipping the file {zip_file} ...")
            with zipfile.ZipFile(zip_file, 'r') as zip_ref:
                zip_ref.extractall(".")
            os.remove(zip_file)
            print("Download and unzip completed.")
            # set the pp and orb path
        
            print("You can set the pseudopotential and orbital path by --pp and --orb, or set the environment variable ABACUS_PP_PATH and ABACUS_ORB_PATH.")
            
        
    
    def run(self,params):
        if params.download_pporb is not None:
            self.download_pporb(params.download_pporb)
            return 0
        
        if params.kpt is not None and len(params.kpt) not in [1, 3]:
            raise ValueError("The kpoint setting should be one or three integers.")
        
        if params.dftu_param is not None:
            dftu_param = self.parse_dftu_param(params.dftu_param)
        else:
            dftu_param = None
            
        if params.init_mag is not None:
            init_mag = self.parse_init_mag(params.init_mag)
        else:
            init_mag = None
        
        pinput = PrepInput(
            files=params.file,
            filetype=params.ftype,
            jobtype=params.jtype,
            pp_path=params.pp,
            orb_path=params.orb,
            input_file=params.input,
            kpt=params.kpt,
            lcao=params.lcao,
            nspin=params.nspin,
            soc=params.soc,
            dftu=params.dftu,
            dftu_param=dftu_param,
            init_mag=init_mag,
            afm=params.afm,
            abacus_command=RECOMMAND_COMMAND,
            machine=RECOMMAND_MACHINE,  # default Bohrium machine type for CPU jobs
            image=RECOMMAND_IMAGE,  # default recommended image for ABACUS jobs 
            copy_pp_orb=params.copy_pp_orb 
        )
        pinput.run()
        return 0

class PrepInput:
    """Prepare the ABACUS inputs for the specified model.
    This class will generate the ABACUS inputs files based on the specified structure files, job type, and other parameters.
    
    Parameters
    ----------
    files : list
        The structure files, should be a list of cif or dpdata supported data
    filetype : str
        The structure type, cif or dpdata supported format
    pp_path : str
        The pseudopotential path, if not specified, will read from the environment variable ABACUS_PP_PATH
    orb_path : str
        The orbital path, if not specified, will read from the environment variable ABACUS_ORB_PATH
    input_file : str
        The template of input file, if not specified, the default input will be generated
    kpt : list
        The kpoint setting, should be a list of three int, will generate a KPT file
    abacus_command : str
        The command to run ABACUS, default is RECOMMAND_COMMAND
    machine : str
        The machine type, default is RECOMMAND_MACHINE, which is the Bohrium machine type for CPU jobs.
    image : str
        The Docker image to use, default is RECOMMAND_IMAGE, which is the recommended image for ABACUS jobs.
    lcao : bool
        Whether to use LCAO basis, default is False, which means using PW basis.
    nspin : int
        The number of spins, can be 1 (no spin), 2 (spin polarized), or 4 (non-collinear spin). Default is 1.
    soc : bool
        Whether to use spin-orbit coupling, if True, nspin should be 4.
    dftu : bool
        Whether to use DFT+U, default is False.
    dftu_param : dict
        The DFT+U parameters, should be a dict like {"Fe": 4, "Ti": 1}, where the key is the element symbol and the value is the U value.
        Value can also be a list of two values, and the first value is the orbital (p, d, f) to apply DFT+U, and the second value is the U value.
        For example, {"Fe": ["d", 4], "O": ["p", 1]} means applying DFT+U to Fe 3d orbital with U=4 eV and O 2p orbital with U=1 eV.
        If dftu is True, but dftu_param is None, the default U values will be used: 4 eV for d orbital elements and 6 eV for f orbital elements.
    init_mag : dict or None
        the initial magnetic moment for magnetic elements, should be a dict like {"Fe": 4, "Ti": 1}, where the key is the element symbol and the value is the initial magnetic moment.
    afm : bool
        Whether to use antiferromagnetic calculation, default is False. If True, half of the magnetic elements will be set to negative initial magnetic moment.
    copy_pp_orb : bool
        Whether to copy the pseudopotential and orbital files to each job directory or link them. Default is False, which means linking the files.
    """
    
    def __init__(self, files: Union[str, List[str], Path, List[Path]],
                 filetype: str, 
                 jobtype: str, 
                 pp_path: Optional[Union[str, Path]]=None, 
                 orb_path: Optional[Union[str, Path]]=None, 
                 input_file: Optional[Union[str, Path]]=None, 
                 kpt: Optional[Tuple[int, int, int]]=None,
                 abacus_command: str=RECOMMAND_COMMAND, 
                 machine: str=RECOMMAND_MACHINE, 
                 image: str=RECOMMAND_IMAGE,
                 lcao: bool=False,
                 nspin: Literal[1,2,4]=1,  # can be 1 or 2 or 4, 1: no spin, 2: spin polarized, 4: non-collinear spin
                 soc: bool =False, # spin-orbit coupling, if True, nspin should be 4 
                 dftu: bool=False,
                 dftu_param: Optional[Dict[str, Union[float, Tuple[Literal["s", 'p', "d"], float]]]]=None,
                 init_mag: Optional[Dict[str, float]] =None,
                 afm: bool = False, 
                 copy_pp_orb: bool = False,
                 ):
        if jobtype not in JOB_TYPES:
            raise ValueError(f"Unsupported job type: {jobtype}.\nSupported job types are {list(JOB_TYPES.keys())}.")
        
        if not isinstance(files, list):
            self.files = [files]
        else:
            self.files = files
        
        # check the parameters
        if nspin not in [1, 2, 4]:
            raise ValueError(f"Invalid nspin value: {nspin}. It should be 1 (no spin), 2 (spin polarized), or 4 (non-collinear spin).")
        if soc and nspin != 4:
            warnings.warn("Spin-orbit coupling is set, but nspin is not 4. Setting nspin to 4 for non-collinear spin calculation.")
            nspin = 4

        self.filetype = filetype
        self.jobtype = jobtype
        self.input_file = input_file
        self.kpt = kpt
        self.abacus_command = abacus_command
        self.machine = machine
        self.image = image
        self.lcao = lcao
        self.nspin = nspin
        self.soc = soc
        self.dftu = dftu
        self.dftu_param = dftu_param
        self.init_mag = init_mag
        self.afm = afm
        self.copy_pp_orb = copy_pp_orb
        
        self.pp_path = pp_path
        self.orb_path = orb_path
        if pp_path is None and os.environ.get("ABACUS_PP_PATH") is not None:
            self.pp_path = os.environ["ABACUS_PP_PATH"]
        if orb_path is None and os.environ.get("ABACUS_ORB_PATH") is not None:
            self.orb_path = os.environ["ABACUS_ORB_PATH"]

        print("Structure files:", self.files)
        print("Structure type:", self.filetype)
        print("Job type:", self.jobtype)
        print("Pseudopotential path:", self.pp_path)
        print("Orbital path:", self.orb_path)
        print("Input file:", self.input_file)
        print("Kpoint:", self.kpt)
        print("LCAO basis:", self.lcao)
        print("Nspin:", self.nspin)
        print("SOC:", self.soc)
        print("DFTU:", self.dftu)
        print("DFTU parameters:", self.dftu_param)
        print("Initial magnetic moment:", self.init_mag)
        print("AFM:", self.afm)
        print("ABACUS command:", self.abacus_command)
        print("Machine:", self.machine)
        print("Image:", self.image)
        print("")
    
    def run(self):
        jobs = gen_stru(self.files, self.filetype, self.pp_path, self.orb_path, tpath=".", copy_pp_orb=self.copy_pp_orb)

        if self.input_file is not None:
            if not os.path.isfile(self.input_file):
                raise ValueError(f"Input file {self.input_file} not found.")
            else:
                input_template = ReadInput(self.input_file)
        else:
            input_template = None
        
        for job in jobs:
            recommand_ecutwfc = jobs[job].get("recommand_ecutwfc", None)
            element = jobs[job]["element"]
            
            input_param = self.generate_input(element)
            input_param = self.update_input(input_param, input_template, recommand_ecutwfc, element)
            
            WriteInput(input_param, os.path.join(job, "INPUT"))
            self.set_init_mag(job, input_param.get("nspin", 1))
            self.write_kpt(job)
        
        job_path = list(jobs.keys())
        if self.jobtype == "band":
            prep_band = PrepBand(job_path, run_command=self.abacus_command)
            _, run_script, extra_files = prep_band.run()
        else:
            run_script_file = "run.sh"
            Path(run_script_file).write_text(self.abacus_command)
            run_script = f"bash {run_script_file}"
            extra_files = [run_script_file]
            

        # generate the parameter setting for abacustest submit
        setting = {
            "save_path": "results",
            "run_dft": {
                "example": job_path,
                "extra_files": extra_files,
                "command": run_script,
                "image": self.image,
                "bohrium": {
                    "scass_type": self.machine,
                    "job_type": "container",
                    "platform": "ali"
                }
            }
        }
        
        setting_file = "setting.json"
        json.dump(setting, open(setting_file, "w"), indent=4)
        print("\nThe inputs are generated in", ", ".join(job_path))
        print(f"Please modify the ABACUS running command in script {extra_files[0]} if needed.")
        print(f"You can modify '{setting_file}', and execute the command 'abacustest submit -p setting.json' to run the abacustest to submit all jobs to bohrium.")
        print(f"Or you can 'cd' to each job path and execute '{extra_files[0]}' to run the job.")
        
        return setting, job_path

    
    def write_kpt(self, job):
        if self.kpt is not None:
            WriteKpt(self.kpt, os.path.join(job, "KPT"))
    
    def set_init_mag(self, job, nspin):
        """
        Set the initial magnetic moment for the job.
        If afm is True, half of the magnetic elements will be set to negative initial magnetic moment.
        """
        if not self.init_mag and not self.afm:
            return
        
        noncolin = nspin == 4
        
        stru_file = os.path.join(job, "STRU")
        if not os.path.isfile(stru_file):
            raise ValueError(f"Structure file {stru_file} not found.")
        stru = AbacusStru.ReadStru(stru_file)
        element = stru.get_element(number=False, total=True)
        
        if self.afm and not self.init_mag:
            mag_element = list(set([e for e in element if e in MAG_ELEMENT]))
            if len(mag_element) == 0:
                raise ValueError(f"No magnetic elements found in the structure {stru_file}, but AFM is set to True. Please set init_mag or remove afm option.")
            else:
                print(f"Setting initial magnetic moment (3 uB) for antiferromagnetic calculation for elements: {mag_element}.")
            init_mag = {e: 3 for e in mag_element}  # default initial magnetic moment is 1 for all magnetic elements
        else:
            init_mag = self.init_mag
        
        atom_mag = [0 for _ in element]
        atommag_idx = {}
        for i, e in enumerate(element):
            if e in init_mag:
                if e not in atommag_idx:
                    atommag_idx[e] = []
                atommag_idx[e].append(i)
                atom_mag[i] = init_mag[e]
        
        if len(atommag_idx) == 0:
            print(f"No magnetic elements found in the structure {stru_file}, skipping initial magnetic moment setting.")
            return
        
        if self.afm:
            # set half of the magnetic elements to negative initial magnetic moment
            print("Setting initial magnetic moment for antiferromagnetic calculation.")
            for e in atommag_idx:
                idx = atommag_idx[e]
                if len(idx) > 1:
                    for i in idx[::2]:
                        atom_mag[i] = -atom_mag[i]
                    print(f"Odd indexed atoms of element {e} are set to negative magnetic moment.")

        if noncolin:
            atom_mag = [[0,0,i] for i in atom_mag]
            print(f"Non-collinear spin calculation, set initial magnetic moment with three components [0, 0, m] for each atom.")
        
        stru.set_atommag(atom_mag)
        stru.write(stru_file)
        
    
    def update_input(self, input_param, input_template, recommand_ecutwfc, element):
        """
        Update the input parameters based on the input template and recommended ecutwfc.
        """
        if input_template is not None: 
            if "calculation" in input_template:
                calculation = input_template["calculation"]
                if calculation != self.jobtype:
                    warnings.warn(f"The calculation type in the input template is {calculation}, but the job type is {self.jobtype}. The input template will be ignored.")
                input_template.pop("calculation")
            input_param.update(input_template)
        else:
            input_template = {}
        
        # only when input_template has not set ecutwfc, and the basis_type is pw, we will reset the ecutwfc based on the recommand_ecutwfc
        if "ecutwfc" not in input_template and input_param.get("basis_type").startswith("pw"):
            if recommand_ecutwfc is not None and list(set(recommand_ecutwfc)) != [None]:
                input_param["ecutwfc"] = max([i for i in recommand_ecutwfc if i is not None])
                print(f"Set ecutwfc to {input_param['ecutwfc']}, element: {element}, recommended ecutwfc: {recommand_ecutwfc}")
        return input_param
    
    def generate_input(self, element):
        """
        Set the special input parameters for the job.
        """
        input_param = JOB_TYPES[self.jobtype].copy()
        pw_basis = input_param.get("basis_type", "pw").startswith("pw")
        
        # LCAO
        if self.lcao:
            input_param.update(LCAO_PARAM)
        
        # NSPIN
        if self.nspin == 2:
            input_param["nspin"] = 2
            input_param["mixing_beta"] = 0.4
            input_param["symmetry"] = 0
            input_param["onsite_radius"] = 3  # to generate the atomic magnetic moment
            input_param["out_mul"] = 1  # to output the mulliken charge
            
        if self.soc or self.nspin == 4:
            if self.nspin != 4:
                warnings.warn("nspin is set to 4 for non-collinear spin calculation.")
            input_param["nspin"] = 4
            input_param["noncolin"] = 1
            input_param["mixing_beta"] = 0.4
            input_param["symmetry"] = -1
            input_param["onsite_radius"] = 3
            input_param["out_mul"] = 1  # to output the mulliken charge
            if self.soc:
                input_param["lspinorb"] = 1
        
        # DFTU
        if self.dftu:
            if self.nspin == 1:
                warnings.warn("DFTU is set, but nspin is 1. We suggest to set nspin to 2 or 4 for DFTU calculation.")
            input_param["dft_plus_u"] = 1
            orbital_corr = []
            hubbard_u = []
            for e in element:
                # set u for element defined in dftu_param
                if self.dftu_param is not None and e in self.dftu_param:
                    if isinstance(self.dftu_param[e], (list,tuple)):
                        orbital = {"p":1, "d":2, "f":3}.get(self.dftu_param[e][0], None)  
                        if orbital is None:
                            raise ValueError(f"Unsupported orbital type {self.dftu_param[e][0]} for element {e}. Supported types are p, d, f.")
                        orbital_corr.append(orbital)
                        hubbard_u.append(self.dftu_param[e][1])
                    else:
                        if e in ELEMENT_DFTU_D:
                            orbital_corr.append(2)
                        elif e in ELEMENT_DFTU_F:
                            orbital_corr.append(3)
                        else:
                            orbital_corr.append(1)
                        hubbard_u.append(self.dftu_param[e])
                # if not define dftu_param, then set u for d and f orbital elements
                elif self.dftu_param is None:
                    if e in ELEMENT_DFTU_D:
                        orbital_corr.append(2)
                        hubbard_u.append(4)  # default U value for d orbital
                    elif e in ELEMENT_DFTU_F:
                        orbital_corr.append(3)
                        hubbard_u.append(6)  # default U value for f orbital
                    else:
                        orbital_corr.append(-1)
                        hubbard_u.append(0)  # no DFTU for other elements
                else:
                    orbital_corr.append(-1)
                    hubbard_u.append(0)  # no DFTU for other elements
            input_param["orbital_corr"] = orbital_corr
            input_param["hubbard_u"] = hubbard_u
            input_param["onsite_radius"] = 3
            if len(set(orbital_corr)) > 1 or orbital_corr[0] != -1:
                if pw_basis:
                    input_param["uramping"] = max(hubbard_u)
                else:
                    input_param["mixing_dmr"] = 1
            
            input_param["mixing_restart"] = 0.001
        
        if input_param.get("basis_type", "pw").startswith("pw") and input_param.get("out_mul", ""):
            # out_mul is only valid for lcao basis
            input_param.pop("out_mul")
            
        if input_param.get("basis_type", "pw").startswith("pw") and input_param.get("nspin",1) == 2 and "onsite_radius" in input_param:
            # onsite_radius is invalid for pw basis with nspin=2
            input_param.pop("onsite_radius")

        return input_param
    


from ..model import Model
import os, glob, json
from . import comm


class CommitTest(Model):
    @staticmethod
    def model_name(): # type: ignore
        '''
        Name of the model, which will be used as the subcommand
        '''
        return "committest"
    
    @staticmethod
    def description(): # type: ignore
        '''
        Description of the model
        '''
        return "Prepare the test on different abacus commit"
    
    @staticmethod
    def prepare_args(parser):
        parser.add_argument("-j","--jobs",default=[],type=str,help="the path of jobs to be tested",action="extend",nargs="*",)
        parser.add_argument("-i","--image",default="intel",type=str,help="the used image. Can be intel/gnu/cuda that means the latest intel/gnu/cuda image, and also can be a self-defined address. Default is intel", )
        parser.add_argument("-c","--commit",default=[],type=str,help="the commits to be tested. Can also be branch or tag name that can be checkouted to using git checkout",action="extend",nargs="*",)
        parser.add_argument("--repo",default="https://github.com/deepmodeling/abacus-develop.git", help="the repository to be tested. Default is https://github.com/deepmodeling/abacus-develop.git")
        parser.add_argument("--clonemax",default=10,help="the max number of commits to be cloned. Default is 10",type=int)
        parser.add_argument("--install", default="cmake -B build -DENABLE_DEEPKS=ON -DENABLE_LIBXC=ON -DENABLE_LIBRI=ON -DENABLE_RAPIDJSON=ON && cmake --build build -j`nproc` && cmake --install build", help="the command to install the abacus. Default same as dockerfile.intel")
        parser.add_argument("--runcommand",default="OMP_NUM_THREADS=1 mpirun -np 16 abacus | tee out.log", help="the command to run the abacus. Default is OMP_NUM_THREADS=1 mpirun -np 16 abacus | tee out.log")
        parser.add_argument("--machine", default="c32_m64_cpu", help="the machine to run the abacus. Default is c32_m64_cpu")
        parser.add_argument("-r", "--run", default=0, help="if run the test. Default is 0.", type=int)
        
    
    def _scripts(self,maxclone,repo,install,command,script_name):
        scripts = """
pwd=$(pwd)
echo -e "[http]\n       proxy = http://ga.dp.tech:8118\n[https]\n     proxy = http://ga.dp.tech:8118\n" > ~/.gitconfig
"""
        scripts += f"""  
cd /      
max_clone={maxclone}
repo="{repo}"
commit=$1
install="{install}"
command="{command}"
tpath=abacus
clone=0

for i in $(seq 1 $max_clone)
do
    git clone --no-checkout {repo} $tpath
    if [ $? -eq 0 ];then
            clone=1
            break
    fi
    
    echo "Clone failed, retry in 5s"
    sleep 5
done

if [ $clone -eq 0 ];then
    echo "Clone failed"
    exit 1
fi

cd $tpath
git checkout $commit
if [ $? -ne 0 ];then
    echo "Checkout failed"
    exit 1
fi
{install}
if [ $? -ne 0 ];then
    echo "Install failed"
    exit 1  
fi

cd $pwd
{command}
        """
        with open(script_name,"w") as f: f.write(scripts)
        
        return scripts
    
        
    def run_prepare(self,params):
        '''
        Parse the parameters and run the prepare process.
        Usually, this step will generate the input files for abacustest submit.
        '''
        jobs = params.jobs
        if len(jobs) == 0:
            print("No jobs are specified.")
            return
        commits = params.commit
        if len(commits) == 0:
            print("No commits are specified.")
            return
        
        self._scripts(params.clonemax,params.repo,params.install,params.runcommand,"run_committest.sh")
        
        real_jobs = comm.get_job_list(jobs)
                    
        setting = {
            "save_path": "results",
            "bohrium_group_name": "committest",
            "run_dft": []
        }
        iamge = {"intel": "registry.dp.tech/deepmodeling/abacus-intel:latest",
                 "gnu": "registry.dp.tech/deepmodeling/abacus-gnu:latest",
                 "cuda": "registry.dp.tech/deepmodeling/abacus-cuda:latest"}
        for commit in commits:
            setting["run_dft"].append({
                "sub_save_path": f"{commit}",
                "example": real_jobs,
                "command": f"bash run_committest.sh {commit}",
                "extra_files": ["run_committest.sh"],
                "image": iamge.get(params.image,params.image),
                "bohrium": {
                    "scass_type": params.machine,
                    "job_type": "container",
                    "platform": "ali"
                }
            })
        comm.dump_setting(setting)
        comm.doc_after_prepare("commit_test",real_jobs,["run_committest.sh","setting.json"],has_prepare=False)
        
        #print("After finish the calculation, you can run below command to do the postprocess:")
        
        if params.run:
            bash_script = "ecutwfc_committest.sh"
            with open(bash_script,"w") as f:
                f.write("abacustest submit -p setting.json\n")
                #f.write("cd results\n")
                #f.write(f"abacustest model {self.model_name()} post -m metrics.json\n")
            os.system(f"bash {bash_script} &")
                    
    
    
    @staticmethod
    def postprocess_args(parser):
        '''
        Add arguments for the postprocess subcommand'''
        pass
    
    
    def run_postprocess(self,params):
        '''
        Parse the parameters and run the postprocess process'''
        pass


import os,sys,glob,re,traceback
from ..resultAbacus import ResultAbacus
from .. import comm
import numpy as np

KS_SOLVER_LIST = ['DA','DS','GE','GV','BP','CG','CU','PE','LA']

class Abacus(ResultAbacus):
    
    @ResultAbacus.register(version="the version of ABACUS")
    def GetVersion(self):
        version = "unknown"
        commit = "unknown"
        if self.JSON:
            version = self.JSON.get("general_info",{}).get("version","unknown")
            commit = self.JSON.get("general_info",{}).get("commit","unknown")
        else:
            if len(self.LOG) > 0:
                for line_idx,line in enumerate(self.LOG):
                    if "WELCOME TO ABACUS" in line:
                        version = line.split()[-1]
                        if version[0].lower() != 'v':
                            print("Unknow version of '%s'" % version)
                        self['version'] = version
                        return
                    elif line[30:36] == "ABACUS":
                        version = line[36:].strip()
                        commit = "unknown"
                        for ii in range(30):
                            if line_idx + ii >= len(self.LOG):
                                break
                            if "Commit:" in self.LOG[line_idx + ii]:
                                commit = re.split(":",self.LOG[line_idx + ii].strip(),maxsplit=1)[-1].strip()
                                break
        self['version'] = version + "(" + commit + ")"
        return
                                              
    @ResultAbacus.register(ncore="the mpi cores",
                           omp_num="the omp cores",)
    def GetNcore(self):
        ncore = omp_num = None
        if self.JSON:
            ncore,_tmp = comm.get_abacus_json(self.JSON,["general_info","mpi_num"])
            omp_num,_tmp = comm.get_abacus_json(self.JSON,["general_info","omp_num"])
        elif self.LOG:    
            for line in self.LOG:
                if "DSIZE =" in line:
                    ncore = int(line.split()[-1])
                    break
                
        self["ncore"] = ncore
        self["omp_num"] = omp_num
    
    @ResultAbacus.register(normal_end="if the job is normal ending")
    def GetNormalEnd(self):
        if len(self.LOG) == 0:
            self['normal_end'] = None
            return
        elif len(self.LOG) > 0 and " Total  Time  :" in self.LOG[-1]:
            self['normal_end'] = True
            return
        else:
            self['normal_end'] = False

            print("Job is not normal ending!!! The latest 10 lines are:")
            if len(self.LOG) < 10:
                print(''.join(self.LOG))
            else:
                print(''.join(self.LOG[-10:]))

    @ResultAbacus.register(INPUT="a dict to store the setting in OUT.xxx/INPUT")
    def GetInputParameter(self):
        def str2intfloat(ii):
            try:
                return int(ii)
            except:
                pass
            try:
                return float(ii)
            except:
                return ii

        INPUTf = os.path.join(self.PATH,"OUT.%s/INPUT" % self.SUFFIX)
        if os.path.isfile(INPUTf):
            with open(INPUTf) as f1:
                input_context = f1.readlines()
        else:
            input_context = self.INPUT
        
        readinput = False
        INPUT = {}
        for i,iline in enumerate(input_context):
            if iline.strip() == 'INPUT_PARAMETERS':
                readinput = True
            elif iline.strip() == '' or iline.strip()[0] in ['#']:
                continue
            elif readinput:
                sline = re.split('[ \t]',iline.split("#")[0].strip(),maxsplit=1)
                if len(sline) == 2:
                    INPUT[sline[0].lower()] = str2intfloat(sline[1].strip())
        self["INPUT"] = INPUT
    
    @ResultAbacus.register(kpt="list, the K POINTS setting in KPT file")
    def GetKptParam(self):
        if len(self.KPT) > 3:
            self["kpt"] = [int(i) for i in self.KPT[3].split()[:3]]
        else:
            self["kpt"] = None

    @ResultAbacus.register(fft_grid = "fft grid for charge/potential")
    def GetGridInfo(self):
        fft_grid = None
        for i,line in enumerate(self.LOG):
            if "fft grid for charge/potential =" in line:
                fft_grid = [float(i.strip()) for i in line.split('=')[1].replace("[","").replace("]","").split(',')]
                break     
        self["fft_grid"] = fft_grid
    
    @ResultAbacus.register(nbase = "number of basis in LCAO")
    def GetLogParamNBase(self): 
        nbase = None
        for i,line in enumerate(self.LOG):
            if "NLOCAL =" in line:
                nbase = int(line.split()[2])
                break
        self['nbase'] = nbase
    
    @ResultAbacus.register(nbands="number of bands",
                           nkstot = "total K point number",
                           ibzk = "irreducible K point number",
                           natom ="total atom number",
                           nelec = "total electron number",
                           nelec_dict = "dict of electron number of each species",
                           point_group="point group",
                           point_group_in_space_group="point group in space group")
    def GetLogParam(self): 
        if self.JSON:
            self["nbands"],_ = comm.get_abacus_json(self.JSON,["init","nband"])
            self["nkstot"],_ = comm.get_abacus_json(self.JSON,["init","nkstot"])
            self["ibzk"],_ = comm.get_abacus_json(self.JSON,["init","nkstot_ibz"])
            self["natom"],_ = comm.get_abacus_json(self.JSON,["init","natom"])
            self["nelec"],_ = comm.get_abacus_json(self.JSON,["init","nelectron"])
            self["nelec_dict"],_ = comm.get_abacus_json(self.JSON,["init","nelectron_each_type"])
            self["fft_grid"],_ = comm.get_abacus_json(self.JSON,["init","fft_grid"])
            self["point_group"],_ = comm.get_abacus_json(self.JSON,["init","point_group"])
            self["point_group_in_space_group"],_ = comm.get_abacus_json(self.JSON,["init","point_group_in_space"])
        else:      
            natom = 0
            nelec = 0
            elec_dict = {}
            point_group = None
            point_group_in_space_group = None
            for i,line in enumerate(self.LOG):
                if "NBANDS =" in line:
                    self['nbands'] = int(line.split()[2])
                elif 'nkstot =' in line:
                    self['nkstot'] = int(line.split()[-1])
                elif 'nkstot_ibz =' in line:
                    self['ibzk'] = int(line.split()[-1])
                elif "            electron number of element" in line:
                    elec_dict[line.split()[4]] = float(line.split()[-1])
                elif 'number of atom for this type =' in line:
                    natom += int(line.split()[-1])
                elif 'total electron number of element' in line:
                    nelec += float(line.split()[-1])
                elif "POINT GROUP =" in line:
                    point_group = line.split("=")[1].strip()
                elif "POINT GROUP IN SPACE GROUP =" in line:
                    point_group_in_space_group = line.split("=")[1].strip()

            self["point_group"] = point_group
            self["point_group_in_space_group"] = point_group_in_space_group
            if natom > 0:
                self["natom"] = natom 
            if nelec > 0:
                self["nelec"] = nelec 
            else:
                self["nelec"] = None

            if elec_dict:
                self["nelec_dict"] = elec_dict
            else:
                self["nelec_dict"] = None

    @ResultAbacus.register(total_mag="total magnetism (Bohr mag/cell)",
                           absolute_mag="absolute magnetism (Bohr mag/cell)",
                           total_mags="total magnetism (Bohr mag/cell) of each ION step",
                           absolute_mags="absolute magnetism (Bohr mag/cell) of each ION step",
                           )
    def GetMagResult(self): 
        total_mags = []
        absolute_mags = []
        tot_mag = None
        abs_mag = None
        ion_step = 1
        for i,line in enumerate(self.LOG):
            if "ION=" in line:
                current_step = int(line.split()[4])
                if current_step > ion_step:
                    if tot_mag is not None:
                        total_mags.append(tot_mag)
                        absolute_mags.append(abs_mag)
                ion_step = current_step 
            elif 'total magnetism (Bohr mag/cell)' in line:
                sline = line.split()
                # if the last three are float, then it is the noncollinear case, else it is collinear case
                try:
                    tot_mag = [float(imag) for imag in sline[-3:]]
                except:
                    tot_mag = float(sline[-1])
            elif 'absolute magnetism' in line:
                abs_mag = float(line.split()[-1])
        
        if tot_mag is not None:    
            self['total_mag'] = tot_mag
            self['absolute_mag'] = abs_mag
            self['total_mags'] = total_mags + [tot_mag]
            self['absolute_mags'] = absolute_mags + [abs_mag]
        else:
            self['total_mag'] = None
            self['absolute_mag'] = None
            self['total_mags'] = None
            self['absolute_mags'] = None
        
    
    @ResultAbacus.register(converge="if the SCF is converged",
                           energy = "the total energy (eV)",
                           energy_ks = "the E_KohnSham, unit in eV",
                           energies = "list of total energy of each ION step",
                           volume = "the volume of cell, in A^3",
                           efermi = "the fermi energy (eV). If has set nupdown, this will be a list of two values. The first is up, the second is down.",
                           energy_per_atom="the total energy divided by natom, (eV)")
    def GetLogResult(self):     
        if self.JSON and self.JSON.get("output") and len(self.JSON.get("output")) > 0:
            output_final = self.JSON.get("output")[-1]
            self["converge"] = output_final.get("scf_converge",None)
            self["energy"] = output_final.get("energy",None)
            self["energies"] = [i.get("energy",None) for i in self.JSON.get("output")]
            self["efermi"] = output_final.get("e_fermi",None)
            cell  = output_final.get("cell",None)
            if cell:
                self["volume"] = abs(np.linalg.det(cell))
            else:
                self["volume"] = None
            if self["natom"] != None and self['energy'] != None:
                self["energy_per_atom"] = self['energy']/self["natom"]
            else:
                self["energy_per_atom"] = None
        else: 
            efermi = None
            converge = None
            energy = None
            energies = []
            energies_ks = []
            volume = None
            for i,line in enumerate(self.LOG):
                if 'charge density convergence is achieved' in line:
                    converge = True
                elif 'convergence has NOT been achieved!' in line or\
                    'convergence has not been achieved' in line:
                    converge = False
                elif "!FINAL_ETOT_IS" in line:
                    energy = float(line.split()[1])
                elif "final etot is" in line:
                    energies.append(float(line.split()[-2]))
                elif "E_KohnSham" in line:
                    energies_ks.append(float(line.split()[-1]))
                elif "Volume (A^3) =" in line:
                    volume = float(line.split()[-1])
                elif 'E_Fermi' in line:
                    if 'E_Fermi_up' in line:
                        if efermi == None:
                            efermi = [None,None]
                        efermi[0] = float(line.split()[-1])
                    elif 'E_Fermi_dw' in line:
                        if efermi == None:
                            efermi = [None,None]
                        efermi[1] = float(line.split()[-1])
                    else:
                        efermi = float(line.split()[-1])

            self["energy"] = energy
            self["converge"] = converge
            self["volume"] = volume
            if energies:
                self["energies"] = energies
            else:
                self["energies"] = None

            if energies_ks:
                self["energy_ks"] = energies_ks[-1]
            else:
                self["energy_ks"] = None

            if efermi != None and isinstance(efermi,list):
                if abs(efermi[0] - efermi[1]) < 1e-6:
                    efermi = efermi[0]
            self["efermi"] = efermi

            if self["natom"] != None and self['energy'] != None:
                self["energy_per_atom"] = self['energy']/self["natom"]
    
    @ResultAbacus.register(force="list[3*natoms], force of the system, if is MD or RELAX calculation, this is the last one",
                           forces = "list of force, the force of each ION step. Dimension is [nstep,3*natom]")
    def GetForceFromLog(self):
        forces = []
        for i in range(len(self.LOG)):
            #i = -1*i - 1
            line = self.LOG[i]
            if 'TOTAL-FORCE (eV/Angstrom)' in line:
                #head_pattern = re.compile(r'^\s*atom\s+x\s+y\s+z\s*$')
                value_pattern = re.compile(r'^\s*[A-Z][a-z]?[1-9][0-9]*\s+[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s+[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s+[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$')
                j = i
                noforce = False
                while not value_pattern.match(self.LOG[j]):
                    j += 1
                    if j >= i + 10:
                        print("Warning: can not find the first line of force")
                        noforce = True
                        break
                if noforce:
                    break
                
                force = []
                while value_pattern.match(self.LOG[j]):
                    force += [float(ii) for ii in self.LOG[j].split()[1:4]]
                    j += 1
                if force: forces.append(force)
        if forces:        
            self['force'] = forces[-1]
            self["forces"] = forces
        else:
            self['force'] = None
            self["forces"] = None
    
    @ResultAbacus.register(stress="list[9], stress of the system, if is MD or RELAX calculation, this is the last one",
                           virial="list[9], virial of the system,  = stress * volume, and is the last one.",
                           pressure="the pressure of the system, unit in kbar.",
                           stresses="list of stress, the stress of each ION step. Dimension is [nstep,9]",
                           virials="list of virial, the virial of each ION step. Dimension is [nstep,9]",
                           pressures="list of pressure, the pressure of each ION step.")
    def GetStessFromLog(self):
        stresses = []
        for i in range(len(self.LOG)):
            #i = -1*i - 1
            line = self.LOG[i]
            if 'TOTAL-STRESS (KBAR)' in line:
                value_pattern = re.compile(r'^\s*[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s+[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s+[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\s*$')
                j = i 
                nostress = False
                # find the first line of stress
                while not value_pattern.match(self.LOG[j]):
                    j += 1
                    if j >= i + 10: # if can not find the first line of force in 10 lines, then stop
                        print("Warning: can not find the first line of force")
                        nostress = True
                        break
                if nostress:
                    break
                stress = []
                while value_pattern.match(self.LOG[j]):
                    stress += [float(ii) for ii in self.LOG[j].split()[:3]]
                    j += 1
                if stress: stresses.append(stress)
        
        virials = []
        pressures = []
        for stress in stresses:
            if self["volume"] != None:
                virial = [i * self["volume"] * comm.KBAR2EVPERANGSTROM3 for i in stress]
                virials.append(virial)
            pressures.append((stress[0] + stress[4] + stress[8]) / 3.0)
            
        if stresses:
            self['stress'] = stresses[-1]
            self["stresses"] = stresses
            self["pressure"] = pressures[-1]
            self["pressures"] = pressures
        else:
            self["stress"] = None
            self["stresses"] = None
            self["pressure"] = None
            self["pressures"] = None

        if virials:
            self['virial'] = virials[-1]
            self["virials"] = virials
        else:
            self['virial'] = None
            self["virials"] = None
        
    @ResultAbacus.register(largest_gradient="list, the largest gradient of each ION step. Unit in eV/Angstrom",
                           largest_gradient_stress="list, the largest stress of each ION step. Unit in kbar")
    def GetLargestGradientFromLog(self):
        lg, lg_stress = None, None
        for line in self.LOG:
            if "Largest gradient in force" in line:
                if lg == None:
                    lg = []
                lg.append(float(line.split()[-2]))
            elif "Largest gradient is" in line:
                if lg == None:
                    lg = []
                lg.append(float(line.split()[-1]))
            elif "Largest gradient in stress" in line:
                if lg_stress == None:
                    lg_stress = []
                lg_stress.append(float(line.split()[-2]))
        self['largest_gradient'] = lg
        self['largest_gradient_stress'] = lg_stress
    
    @ResultAbacus.register(k_coord="list, the direct k point coordinates in the BZ",)
    def GetKCoordFromLog(self):
        coord = []
        nkstot = self["nkstot"]
        for i, line in enumerate(self.LOG):
            if "K-POINTS DIRECT COORDINATES" in line:
                for j in range(i+2, i+2+nkstot):
                    coord.append([float(k) for k in self.LOG[j].split()[1:4]])
                break
        if len(coord) > 0:
            self['k_coord'] = coord
        else:
            self['k_coord'] = None
    
    @ResultAbacus.register(band = "Band of system. Dimension is [nspin,nk,nband].",
                           band_weight = "Band weight of system. Dimension is [nspin,nk,nband]."
                           )
    def GetBand(self): 
        band_file = os.path.join(self.PATH,f"OUT.{self.SUFFIX}/eig.txt") # in new version the band info is in eig.txt
        if os.path.isfile(band_file):
            band = []
            weight = []
            with open(band_file) as f: lines = f.readlines()
            for line in lines[2:]:
                if line.strip() == "":
                    continue
                elif line.startswith(" spin="):
                    spin_idx = int(line[6]) # start with 1
                    if spin_idx > len(band):
                        band.append([])
                        weight.append([])
                    band[-1].append([])
                    weight[-1].append([])
                else:
                    sline = line.split()
                    if len(sline) == 3:
                        band[-1][-1].append(float(sline[1]))
                        weight[-1][-1].append(float(sline[2]))
            self['band'] = band
            self['band_weight'] = weight
        else:
            nband = self['nbands']  
            if self["ibzk"] != None:    
                nk = self['ibzk']
            elif self["nkstot"] != None:
                nk = self["nkstot"]
            else:
                nk = None

            if nband == None or nk == None:
                print("no nbands or ibzk, and skip the catch of band info")
                return

            band = None
            band_weight = None
            for i,line in enumerate(self.LOG):
                if 'STATE ENERGY(eV) AND OCCUPATIONS' in line:
                    nspin = int(line.split()[-1])
                    if nspin == 4: nspin=1  # for nspin4, only total band is output
                    band = []
                    band_weight = []
                    for ispin in range(nspin):
                        band.append([])
                        band_weight.append([])
                        for k in range(nk):
                            band[-1].append([])
                            band_weight[-1].append([])
                            for m in range(nband):
                                ni = ((nband+2)*nk + 1) * ispin + (nband+2)*k + nspin + m + 1 + i
                                eband = float(self.LOG[ni].split()[1])
                                wband = float(self.LOG[ni].split()[2])
                                band[-1][-1].append(eband)
                                band_weight[-1][-1].append(wband)
                    break
            self['band'] = band
            self['band_weight'] = band_weight
    
    @ResultAbacus.register(band_plot="Plot the band structure. Return the file name of the plot.")
    def PlotBandFromLog(self):  
        band = self['band']  
        efermi = self['efermi']
        if band == None:
            print("no band, and skip the plot of band")
            self['band_plot'] = None
            return
        band_plot = os.path.join(self.PATH,"band.png")
        comm.plot_band(band, band_plot, efermi)
        self['band_plot'] = band_plot     
    
    @ResultAbacus.register(e_bandgap= "The band gap outputted in running_xxx.log, unit in eV")
    def GetEBandGapFromLog(self):
        ebandgap = None 
        for line in self.LOG[::-1]:
            if "E_bandgap" in line:
                ebandgap = float(line.split()[-1])
                break
        self['e_bandgap'] = ebandgap
    
    @ResultAbacus.register(band_gap = "band gap of the system")
    def GetBandGapFromLog(self):
        def ErrorReturn(strinfo):
            print(f"WARNING: {strinfo}, skip the catch of band gap info")
            self['band_gap'] = None
            return
        
        band = self['band']
        band_weight = self['band_weight']
        if band == None or band_weight == None:
            ErrorReturn("no band")
            return
        
        efermi = self['efermi']
        if efermi == None:
            ErrorReturn("no efermi")
            return
        
        self['band_gap'] = comm.cal_band_gap(band,efermi)
    '''
        
    @ResultAbacus.register(band_gap = "band gap of the system")
    def GetBandGapFromLog(self):
        def ErrorReturn(strinfo):
            print("WARNING: %s, skip the catch of band gap info" % strinfo)
            self['band_gap'] = None
            return
        
        nband = self['nbands']  
        if self["ibzk"] != None:    
            nk = self['ibzk']
        elif self["nkstot"] != None:
            nk = self["nkstot"]
        else:
            nk = None
            
        if nband == None or nk == None:
            ErrorReturn("no nbands or ibzk")
            return
        nelec = self['nelec']
        if nelec == None:
            ErrorReturn("no nelec")
            return
        occu_band = int(nelec/2)
        if occu_band == 0:
            ErrorReturn("no occu_band")
            return
        if occu_band >= nband:
            ErrorReturn("occu_band >= nband")
            return
                
        band_gap = None
        #print("nelec:",nelec,"occu_band:",occu_band,"nband:",nband,"nk:",nk)
        for i,line in enumerate(self.LOG):
            if 'STATE ENERGY(eV) AND OCCUPATIONS' in line:
                nspin = int(line.split()[-1])
                if nspin not in [1,2]:
                    ErrorReturn("NOT SUPPORT FOR NSPIN=%d now" % nspin)
                    return

                totalcb = None
                totalvb = None
                for ispin in range(nspin):
                    cb = None
                    vb = None
                    for k in range(nk):
                        eband1 = float(self.LOG[((nband+2)*nk + 1) * ispin + (nband+2)*k + nspin + occu_band + i].split()[1])  # the highest occupied band
                        eband2 = float(self.LOG[((nband+2)*nk + 1) * ispin + (nband+2)*k + nspin + occu_band + 1 + i].split()[1])  # the lowest unoccupied band
                        if cb == None or eband1 > cb:
                            cb = eband1
                        if vb == None or eband2 < vb:
                            vb = eband2
                        #print("k:",k,"eband1:",eband1,"eband2:",eband2,"cb:",cb,"vb:",vb)
                    if totalcb == None or (cb != None and totalcb < cb): totalcb = cb
                    if totalvb == None or (vb != None and totalvb > vb): totalvb = vb
                    #print("ispin:",ispin,"totalcb:",totalcb,"totalvb:",totalvb)
                if totalcb == None or totalvb == None:
                    band_gap = None
                else:
                    band_gap = totalvb-totalcb
                    if band_gap < 0: band_gap = 0
                break
                
        self['band_gap'] = band_gap
    '''
    
    @ResultAbacus.register(total_time="the total time of the job",
                           stress_time="the time to do the calculation of stress",
                           force_time = "the time to do the calculation of force",
                           scf_time = "the time to do SCF",
                           scf_time_each_step = "list, the time of each step of SCF",
                           step1_time = "the time of 1st SCF step",
                           scf_steps = "the steps of SCF")
    def GetTimeFromOutput(self):
        # first, check if self.time is already exist
        if self.TIME:
            total_time = self.GetTime("total",None)[0]
            # if PW basis, stress time = Stress_PW/cal_stress, force_time = cal_force_nl
            # if lcao basis, stress time = getForceStress, force_time = None
            stress_time_pw = self.GetTime("Stress_PW","cal_stress")[0]
            stress_time_lcao = self.GetTime("Force_Stress_LCAO","getForceStress")[0]
            force_time_pw = self.GetTime("Forces","cal_force_nl")[0]
            force_time_lcao = None
            if stress_time_pw != None:
                stress_time = stress_time_pw
            else:
                stress_time = stress_time_lcao
                
            if force_time_pw != None:
                force_time = force_time_pw
            else:
                force_time = force_time_lcao
        else:
            stress_time = None
            force_time = None
            total_time = None
            for i,line in enumerate(self.OUTPUT):
                if line[23:28] == 'total' or (len(line.split()) == 5 and line.split()[0] == 'total'):  # old version or new version
                    total_time = float(line.split()[1])
                elif line[23:33] == 'cal_stress' or (len(line.split()) == 6 and line.split()[0] == 'cal_stress'): # old version or new version
                    stress_time = float(line.split()[-5])
                elif line[23:35] == 'cal_force_nl' or (len(line.split()) == 6 and line.split()[0] == 'cal_force_nl'):
                    force_time = float(line.split()[-5])
                elif line[23:37] == 'getForceStress' or (len(line.split()) == 6 and line.split()[0] == 'getForceStress'):
                    stress_time = float(line.split()[-5])
                    force_time = None
                    
        self["total_time"] = total_time
        self['stress_time'] = stress_time
        self['force_time'] = force_time

        scftime = []
        for i,line in enumerate(self.OUTPUT):
            if line[1:5] == 'ITER':
                for j in range(i+1,len(self.OUTPUT)):
                    if self.OUTPUT[j][1:3] in KS_SOLVER_LIST:
                        scftime.append(float(self.OUTPUT[j].split()[-1]))
                break
        if len(scftime) > 0:
            self['scf_time'] = np.array(scftime).sum()
            self['step1_time'] = scftime[0]
            self['scf_steps'] = len(scftime)
            self['scf_time_each_step'] = scftime

    @ResultAbacus.register(
                           atom_mag = "list, the magnization of each atom. Only the last ION step. Calculated by atomic orbital projection.",
                           atom_mags= "list atom mag of each ION step. Calculated by atomic orbital projection.",
                         )
    def GetAtomMagOrbital(self):
        self["atom_mags"] = self["ds_mags"]
        self["atom_mag"] = self["ds_mag"]
    
    @ResultAbacus.register(atom_orb_mag = "list of atomic orbital magnization. Only the last ION step. Calculated by atomic orbital projection.",
                            atom_orb_mags = "list of atomic orbital magnization of each ION step. Calculated by atomic orbital projection.",
                            atom_elec = "list of atomic electron. Electron calculated by atomic orbital projection.",
                            atom_elecs="list of atomic electron of each ION step. Electron calculated by atomic orbital projection.",
                            atom_orb_elec="list of atomic orbital electron. Electron calculated by atomic orbital projection.",
                            atom_orb_elecs="list of atom_orb_elec of each ION step. Electron calculated by atomic orbital projection.",
                          )
    def GetAtomMagOrbital(self):
        """Get electrion and mag from orbital charge analysis.
        """
        
        '''
-------------------------------------------------------------------------------------------
Orbital Charge Analysis      Charge         Mag(x)         Mag(y)         Mag(z)
-------------------------------------------------------------------------------------------
Fe1
                   s         1.0799        -0.0000         0.0000         0.0034
                   p         5.9969        -0.0000         0.0000         0.0004
                   d         6.5446        -0.0039         0.0013         3.0690
                 Sum        13.6214        -0.0039         0.0013         3.0729
Fe2
                   s         1.0827         0.0000        -0.0000         0.0040
                   p         5.9969         0.0000        -0.0000         0.0004
                   d         6.6497         0.0025        -0.0002         2.9733
                 Sum        13.7294         0.0025        -0.0002         2.9776
-------------------------------------------------------------------------------------------        
        '''
        
        atom_mags = []
        atom_orb_mags = []
        atom_elecs = []
        atom_orb_elecs = []
        for i, line in enumerate(self.LOG):
            if "Orbital Charge Analysis      Charge" in line:
                atom_mag = []
                atom_orb_mag = []
                atom_elec = []
                atom_orb_elec = []
                j = i + 2
                while "-------" not in self.LOG[j]:
                    sline = self.LOG[j].split()
                    if len(sline) == 1:
                        atom_orb_mag.append([])
                        atom_orb_elec.append([])
                    else:
                        if "Sum" in sline:
                            atom_elec.append(float(sline[1]))
                            atom_mag.append([float(ii) for ii in sline[2:]])
                        else:
                            atom_orb_mag[-1].append(float(sline[1]))
                            atom_orb_elec[-1].append([float(ii) for ii in sline[2:]])
                    j += 1
                atom_mags.append(atom_mag)
                atom_orb_mags.append(atom_orb_mag)
                atom_elecs.append(atom_elec)
                atom_orb_elecs.append(atom_orb_elec)
        
        if len(atom_orb_mags) == 0:
            self["atom_orb_mags"] = None
            self["atom_orb_mag"] = None
            self["atom_elecs"] = None
            self["atom_elec"] = None
            self["atom_orb_elecs"] = None
            self["atom_orb_elec"] = None
        else:
            self["atom_orb_mags"] = atom_orb_mags
            self["atom_orb_mag"] = atom_orb_mags[-1]
            self["atom_elecs"] = atom_elecs
            self["atom_elec"] = atom_elecs[-1]
            self["atom_orb_elecs"] = atom_orb_elecs
            self["atom_orb_elec"] = atom_orb_elecs[-1]
    
    @ResultAbacus.register(atom_mag_mul = "list of mulliken magnization of each atom. Only the last ION step.",
                           atom_mags_mul="list of atom_mag_mul of each ION step.",
                           atom_elec_mul="list of mulliken atomic electron.",
                           atom_orb_elec_mul="list of mulliken atomic orbital electron.",
                           )
    def GetAtomMagMul(self):
        mullikenf = os.path.join(os.path.split(self.LOGf)[0],"mulliken.txt")
        if not os.path.isfile(mullikenf):
            self['atom_mag_mul'] = None
            self["atom_mags_mul"] = None
            self["atom_elec"] = None
            return
        
        atom_mag = []
        atom_elec = []
        with open(mullikenf) as f1: lines = f1.readlines()
        for idx,line in enumerate(lines):
            if line[:5] == "STEP:":
                atom_mag.append([])
            elif "Total Magnetism on atom" in line:
                if "(" in line and ")" in line:
                    # for nspin = 4
                    sline = line.split("(")[1].split(")")[0].split(",")
                    if len(sline) == 3:
                        atom_mag[-1].append([float(sline[0]),float(sline[1]),float(sline[2])])
                else:
                    atom_mag[-1].append(float(line.split()[-1]))
            elif "Zeta of" in line:
                two_spin = True if "Spin 2" in line else False
                atom_elec.append([])
                j = idx + 1
                while j < len(lines) and lines[j].strip() != "":
                    if "Zeta of" in lines[j]:
                        break
                    if "sum over m+zeta" in lines[j]:
                        atom_elec[-1].append([float(lines[j].split()[3])])
                        if two_spin:
                            atom_elec[-1][-1].append(float(lines[j].split()[4]))
                    j += 1
        
        if len(atom_mag) == 0:
            self['atom_mags_mul'] = None
            self["atom_mag_mul"] = None
        else:
            self['atom_mags_mul'] = atom_mag
            self["atom_mag_mul"] = atom_mag[-1]
        
        if not atom_elec:
            self["atom_elec_mul"] = None
            self["atom_orb_elec_mul"] = None
        else:
            self["atom_orb_elec_mul"] = atom_elec
            atom_elec = []
            for i in self["atom_orb_elec_mul"]:
                t = 0
                for j in i:
                    t += sum(j)
                atom_elec.append(t)
            self["atom_elec_mul"] = atom_elec
            
    '''
    repeat to charge, not used anymore
    @ResultAbacus.register(atom_mag_uorb="list of a dict, the magnization of each atom U orbitals. Only the last SCF step.",
                           atom_elec_uorb = "list of a dict with keys are atom index, atom label, and electron of orbital adding U.")
    def GetAtomMag(self):
        atom_mag_u = None
        atom_elec_u = None
        if self.LOG:
            u_block = []
            start_line = end_line = None
            for i in  range(len(self.LOG)):
                i = -1*i - 1
                if "//=========================L(S)DA+U===========================//" in self.LOG[i]:
                    start_line = i + 1
                    break
                elif "//=======================================================//" in self.LOG[i]:
                    end_line = i
            print(start_line,end_line)
            if None not in [start_line, end_line]:
                u_block = self.LOG[start_line:end_line]
            else:
                print("Can not find the L(S)DA+U block")
                self['atom_mag_uorb'] = None
                self["atom_elec_uorb"] = None
                return
            
            labels = self["label"]
            atom_mag_u = []
            atom_elec_u = []
            for i,line in enumerate(u_block):
                if "atoms" in line:
                    atom_index = int(line.split()[-1])
                    atom_label = None if not labels else labels[atom_index]
                    atom_mag_u.append({"index":atom_index,"label":atom_label,"mag":None})
                    atom_elec_u.append({"index":atom_index,"label":atom_label,"elec":None})
                elif "eigenvalues" in line:
                    if atom_elec_u[-1]["elec"] == None:
                        atom_elec_u[-1]["elec"] = []
                    atom_elec_u[-1]["elec"].append(float(u_block[i+1].split()[-1]))
                elif "atomic mag:" in line:
                    atom_mag_u[-1]["mag"] = float(line.split()[-1])
        self['atom_mag_uorb'] = atom_mag_u
        self['atom_elec_uorb'] = atom_elec_u
    '''
    
    @ResultAbacus.register(drho="[], drho of each scf step",
                           drho_last="drho of the last scf step")
    def GetDrho(self):
        drho = []
        for line in self.LOG:
            if "Density error is" in line:
                drho.append(float(line.split()[-1]))
        
        if len(drho) == 0:
            self['drho'] = None
        else:
            self['drho'] = drho
            self["drho_last"] = drho[-1]
    
    @ResultAbacus.register(denergy="[], denergy of each scf step",
                           denergy_last="denergy of the last scf step")
    def GetDenergy(self):
        denergy = None
        if self.OUTPUT:
            for i,line in enumerate(self.OUTPUT):
                if "ITER" in line and "ETOT/eV" in line and "EDIFF/eV" in line and "DRHO" in line and "TIME/s" in line:
                    denergy = []
                    ncol = len(line.split())
                    ediff_idx = line.split().index("EDIFF/eV")
                    j = i + 1
                    while j < len(self.OUTPUT):
                        if "----------------------------" in self.OUTPUT[j]:
                            break
                        if self.OUTPUT[j][1:3] in KS_SOLVER_LIST and len(self.OUTPUT[j].split()) == ncol:
                            denergy.append(float(self.OUTPUT[j].split()[ediff_idx]))
                        j += 1

        self["denergy"] = denergy
        if denergy != None:
            self["denergy_last"] = denergy[-1]
        else:
            self["denergy_last"] = None
    
    @ResultAbacus.register(denergy_womix="[], denergy (calculated by rho without mixed) of each scf step",
                           denergy_womix_last="float, denergy (calculated by rho without mixed) of last scf step")
    def GetDenergyWOMIX(self):
        des = []
        if self.LOG:
            for line in self.LOG:
                if "DeltaE_womix" in line:
                    des.append(float(line.split()[-2]))
        if len(des) == 0:
            self["denergy_womix"]    = None
            self["denergy_womix_last"] = None
        else:
            des[0] = 0
            self["denergy_womix"]    = des
            self["denergy_womix_last"] = des[-1]


    @ResultAbacus.register(lattice_constant="a list of six float which is a/b/c,alpha,beta,gamma of cell. If has more than one ION step, will output the last one.",
                           lattice_constants="a list of list of six float which is a/b/c,alpha,beta,gamma of cell",
                           cell = "[[],[],[]], two-dimension list, unit in Angstrom. If is relax or md, will output the last one.",
                           cells = "a list of [[],[],[]], which is a two-dimension list of cell vector, unit in Angstrom.",
                           cell_init = "[[],[],[]], two-dimension list, unit in Angstrom. The initial cell",
                           coordinate = "[[],..], two dimension list, is a cartesian type, unit in Angstrom. If is relax or md, will output the last one",
                           coordinate_init = "[[],..], two dimension list, is a cartesian type, unit in Angstrom. The initial coordinate",
                           element = "list[], a list of the element name of all atoms",
                           label = "list[], a list of atom label of all atoms",
                           element_list = "same as element",
                           atomlabel_list = "same as label")
    def GetCell(self): 
        lc = 1   
        for line in self.LOG:
            if "lattice constant (Angstrom)" in line:
                lc = float(line.split()[-1])
                break
        
        cells = []
        for i in range(len(self.LOG)): 
            iline = -i - 1 
            line = self.LOG[iline]
            if "Lattice vectors: (Cartesian coordinate: in unit of a_0)" in line:
                icell = []
                for k in range(1, 4):
                    icell.append([float(x) * lc for x in self.LOG[iline + k].split()[0:3]])
                cells.append(icell)
                break
                    
        if cells:
            self['cells'] = cells
            self['cell'] = cells[-1]
            # calculate a/b/c,alpha,beta,gamma
            lattice_constants = []
            for cell in cells:
                a = np.linalg.norm(cell[0])
                b = np.linalg.norm(cell[1])
                c = np.linalg.norm(cell[2])
                alpha = np.arccos(np.dot(cell[1],cell[2])/(b*c))*180/np.pi
                beta = np.arccos(np.dot(cell[0],cell[2])/(a*c))*180/np.pi
                gamma = np.arccos(np.dot(cell[0],cell[1])/(a*b))*180/np.pi
                lattice_constants.append([a,b,c,alpha,beta,gamma])
            self['lattice_constants'] = lattice_constants
            self['lattice_constant'] = lattice_constants[-1]
        else:
            self["cells"] = None
            self["cell"] = None
            self['lattice_constants'] = None
            self['lattice_constant'] = None
        
        cell_init = None
        for i in range(len(self.LOG)): 
            iline = i 
            line = self.LOG[iline]
            if "Lattice vectors: (Cartesian coordinate: in unit of a_0)" in line:
                cell_init = []
                for k in range(1, 4):
                    cell_init.append([float(x) * lc for x in self.LOG[iline + k].split()[0:3]])
                break
        self['cell_init'] = cell_init
        
        coordinate = None
        coordinate_init = None
        natom = self['natom']
        if cell != None and natom != None:
            try:
                for i in range(len(self.LOG)): 
                    iline = -i - 1 
                    line = self.LOG[iline]
                    if len(line.split()) >= 2 and line.split()[1] == "COORDINATES":  
                        coordinate = []
                        if line.split()[0] == "DIRECT":
                            for k in range(2, 2 + natom):
                                coordinate.append([float(x) for x in self.LOG[iline + k].split()[1:4]])
                            coordinate = np.array(coordinate).dot(np.array(cell)).tolist()
                        elif line.split()[0] == "CARTESIAN":
                            for k in range(2, 2 + natom):
                                coordinate.append([float(x) for x in self.LOG[iline + k].split()[1:4]])
                        else:
                            print("Unrecongnized coordinate type: %s" % (line))   
                        break
                self['coordinate'] = coordinate  
            except:
                traceback.print_exc()
                self['coordinate'] = None 
            
            try:
                for i in range(len(self.LOG)): 
                    iline = i
                    line = self.LOG[iline]
                    if len(line.split()) >= 2 and line.split()[1] == "COORDINATES":  
                        coordinate_init = []
                        if line.split()[0] == "DIRECT":
                            for k in range(2, 2 + natom):
                                coordinate_init.append([float(x) for x in self.LOG[iline + k].split()[1:4]])
                            coordinate_init = np.array(coordinate_init).dot(np.array(cell)).tolist()
                        elif line.split()[0] == "CARTESIAN":
                            for k in range(2, 2 + natom):
                                coordinate_init.append([float(x) for x in self.LOG[iline + k].split()[1:4]])
                        else:
                            print("Unrecongnized coordinate type: %s" % (line))   
                        break
                self['coordinate_init'] = coordinate_init 
            except:
                traceback.print_exc()
                self['coordinate_init'] = None   
        else:
            print("No cell or natom, skip the catch of coordinate info")
            self['coordinate'] = None
            self['coordinate_init'] = None
              
        element_list = None
        atomlabel_list = None
        for line in self.LOG:
            if "atom label =" in line:
                label = line.split()[-1]
                element = label
                while element[-1].isdigit():
                    element = element[:-1]
            elif "number of atom for this type =" in line:
                if element_list == None: element_list = []
                if atomlabel_list == None: atomlabel_list = []
                natom = int(line.split()[-1])
                atomlabel_list += natom * [label]
                element_list += natom * [element]
                if len(atomlabel_list) >= self['natom']:
                    break
        self["element"] = element_list
        self["label"] = atomlabel_list
        self['element_list'] = element_list
        self['atomlabel_list'] = atomlabel_list
    
    @ResultAbacus.register(pdos="dict, where keys are 'energy' and 'orbitals', and value of 'orbitals' is a dict of (index,species,l,m,z,data)",
                           )
    def GetPDOS(self): 
        pdos_file = os.path.join(self.PATH,f"OUT.{self.SUFFIX}","PDOS")
        print("PDOS file path:",pdos_file)
        pdos = None
        if os.path.isfile(pdos_file):
            import xml.etree.ElementTree as ET
            tree = ET.parse(pdos_file)
            root = tree.getroot()
            nspin = int(root.find('nspin').text)
            energy = [float(i) for i in root.find('energy_values').text.split()]

            ne = len(energy)

            all_orbitals = []
            for iorb in root.findall('orbital'):
                data = [[] for i in range(nspin)]
                for i in iorb.find('data').text.split("\n"):
                    for j,jj in enumerate(i.split()):
                        if j == 1:
                            data[j].append(-1*float(jj))
                        else:
                            data[j].append(float(jj))
                if len(data[0]) != ne:
                    print("WARNING: PDOS len(data[0]) != ne")      
                all_orbitals.append({
                    "index": int(iorb.get('index')),
                    "atom_index": int(iorb.get('atom_index')),
                    "species": iorb.get('species'),
                    "l": int(iorb.get('l')),
                    "m": int(iorb.get('m')),
                    "z": int(iorb.get('z')),
                    "data": data
                })
            pdos = {"energy":energy,"orbitals":all_orbitals}
        self["pdos"] = pdos

class AbacusRelax(ResultAbacus):
    @ResultAbacus.register(relax_converge="if the relax is converged")
    def GetRelaxConverge(self):
        #need read self.LOG
        if self.LOG:
            for i in range(len(self.LOG)):
                line = self.LOG[-i-1]
                if "Relaxation is converged!" in line:
                    self["relax_converge"] = True
                    return
                elif "Relaxation is not converged yet!" in line:
                    self["relax_converge"] = False
                    return
                elif "Ion relaxation is not converged yet" in line or \
                    "Lattice relaxation is not converged yet" in line:
                    self["relax_converge"] = False
                    return
                elif "Lattice relaxation is converged!" in line or \
                    "Ion relaxation is converged!" in line:
                    self["relax_converge"] = True
                    return
        self["relax_converge"] = None
    
    @ResultAbacus.register(relax_steps= "the total ION steps")
    def GetRelaxSteps(self):
        #need read self.LOG
        if self.LOG:
            for i in range(len(self.LOG)):
                line = self.LOG[-i-1]
                if "ALGORITHM --------------- ION=" in line:
                    index_ben = line.index("ION=") + 4
                    index_end = line.index("ELEC")
                    self["relax_steps"] = int(line[index_ben:index_end])
                    return
                elif "STEP OF RELAXATION :" in line:
                    self["relax_steps"] = int(line.split()[-1])
                    return
                elif " STEP OF ION RELAXATION : " in line:
                    self["relax_steps"] = int(line.split()[-1])
                    return
        self["relax_steps"] = None


class AbacusDeltaSpin(ResultAbacus):
    
    @ResultAbacus.register(ds_lambda_step="a list of DeltaSpin converge step in each SCF step",
                           ds_lambda_rms="a list of DeltaSpin RMS in each SCF step",
                           ds_time="a list of the total time of inner loop in deltaspin for each scf step.")
    def GetDSLog(self):
        """Get metrcis of DeltaSpin from the log file.
        """
        
        '''
Step (Outer -- Inner) =  16 -- 1           RMS = 1.057e-07
Step (Outer -- Inner) =  16 -- 2           RMS = 1.355e-07
Step (Outer -- Inner) =  16 -- 3           RMS = 1.176e-07
Step (Outer -- Inner) =  16 -- 4           RMS = 9.760e-08

        '''
        lambda_step = None 
        lambda_rms = None
        ds_time = None
        if self.OUTPUT:
            scf_step = []
            lambda_step = []
            lambda_rms = []
            for idx, i in enumerate(self.OUTPUT):
                if "Step (Outer -- Inner) =" in i:
                    ss = int(i.split()[5])
                    lambdas = int(i.split()[7])
                    rms = float(i.split()[10])
                    if ss not in scf_step:
                        scf_step.append(ss)
                        lambda_step.append(lambdas)
                        lambda_rms.append(rms)
                    else:
                        lambda_step[-1] = lambdas
                        lambda_rms[-1] = rms
                elif "Meet convergence criterion" in i and "Total TIME(s) =" in i:
                    if ds_time is None:
                        ds_time = []
                    ds_time.append(float(i.split()[-1]))

        self["ds_lambda_step"] = lambda_step
        self["ds_lambda_rms"] = lambda_rms   
        self["ds_time"] = ds_time
        
    @ResultAbacus.register(ds_mag="a list of list, each element list is for each atom. Unit in uB",
                           ds_mags="list of ds_mag of each ION step. Unit in uB",
                           ds_mag_force="a list of list, each element list is for each atom. Unit in eV/uB",
                           ds_mag_forces="list of ds_mag_force of each ION step. Unit in eV/uB")
    def GetDSOutput(self):
        '''
===============================================================================
 DA13     1.98e+00  -7.38e-10   3.44e+00   4.88e+00  -6.81997451e+03   1.78754645e-06   2.5764e-09 108.32
===============================================================================
Inner optimization for lambda begins ...
Covergence criterion for the iteration: 1e-08
initial lambda (eV/uB):
ATOM      1         0.1482429405         0.0000000026         0.0156735253
ATOM      2        -0.0605498198         0.0000000025         0.1362180540
initial spin (uB):
ATOM      1         0.0000367975        -0.0000000063         2.4319892847
ATOM      2         2.1061545741        -0.0000000061         1.2160301143
target spin (uB):
ATOM      1         0.0000000000         0.0000000000         2.4320000000
ATOM      2         2.1061737820         0.0000000000         1.2160000000
Step (Outer -- Inner) =  13 -- 1           RMS = 3.70452e-05     TIME(s) = 24.4197
Step (Outer -- Inner) =  13 -- 2           RMS = 3.15459e-06     TIME(s) = 16.202
Step (Outer -- Inner) =  13 -- 3           RMS = 8.99985e-07     TIME(s) = 16.2116
Step (Outer -- Inner) =  13 -- 4           RMS = 1.42306e-07     TIME(s) = 16.2354
Step (Outer -- Inner) =  13 -- 5           RMS = 3.15143e-08     TIME(s) = 16.2353
Meet convergence criterion ( < 3.70452e-08 ), exit.       Total TIME(s) = 89.304
after-optimization spin (uB): (print in the inner loop):
ATOM      1        -0.0000000053        -0.0000000001         2.4319999813
ATOM      2         2.1061737840        -0.0000000001         1.2159999599
after-optimization lambda (eV/uB): (print in the inner loop):
ATOM      1         0.1482675982        -0.0000000006         0.0156621842
ATOM      2        -0.0605667962        -0.0000000007         0.1362365586
Inner optimization for lambda ends.  
...
        '''
        ds_mag1 = []
        ds_mag2 = []
        mag_forces = []

        if self.LOG:
            natom = self["natom"]
            for idx, i in enumerate(self.LOG):
                if "Total Magnetism (uB)" in i:
                    ds_mag = []
                    for j in range(natom):
                        if len(self.LOG[idx+j+2].split()) in [2,4]:
                            ds_mag.append([float(ii) for ii in self.LOG[idx+j+2].split()[1:]])
                    ds_mag1.append(ds_mag)
                elif "Orbital Charge Analysis      Charge" in i:
                    ds_mag = []
                    j = idx +2
                    while "-------" not in self.LOG[j]:
                        if "Sum" in self.LOG[j]:
                            ds_mag.append([float(ii) for ii in self.LOG[j].split()[2:]])   
                        j += 1 
                    ds_mag2.append(ds_mag)    
                elif "Magnetic force (eV/uB)" in i:
                    mag_force = []
                    for j in range(natom):
                        if len(self.LOG[idx+j+2].split()) in [2,4]:
                            mag_force.append([float(ii) for ii in self.LOG[idx+j+2].split()[1:]])
                    mag_forces.append(mag_force)

        if len(ds_mag1) == 0 and len(ds_mag2) == 0:
            self["ds_mag"] = None
            self["ds_mags"] = None
        elif len(ds_mag1) > 0:
            self["ds_mag"] = ds_mag1[-1]
            self["ds_mags"] = ds_mag1
        else:
            self["ds_mag"] = ds_mag2[-1]
            self["ds_mags"] = ds_mag2
        if len(mag_forces) == 0:
            self["ds_mag_force"] = None
            self["ds_mag_forces"] = None
        else:
            self["ds_mag_forces"] = mag_forces
            self["ds_mag_force"] = mag_forces[-1]

        
class AbacusMemory(ResultAbacus):
    
    @ResultAbacus.register(mem_vkb="the memory of VNL::vkb, unit it MB",
                           mem_psipw="the memory of PsiPW, unit it MB",)
    def GetMemory(self):
        if self.LOG:
            for line in self.LOG:
                if "Warning_Memory_Consuming allocated:  VNL::vkb" in line:
                    self["mem_vkb"] = float(line.split()[-2])
                elif "Warning_Memory_Consuming allocated:  Psi_PW" in line:
                    self["mem_psipw"] = float(line.split()[-2])
                    

         





from . import (comm_class,
               comm_class_exampleSource,
               comm_class_predft,
               comm_class_rundft,
               comm_class_postdft,
               comm_class_metrics,
               comm_class_prepare,
               comm_func)

def ReadSetting(logs:comm_class.myLog,opts,work_path,download_path):
    """
    {
        "config":{},
        "pre_dft":{
            "image":
            "bohrium":{"scass_type","job_type","platform"},
            "example":[],
            "extra_files":[],
            "command":
            "work_directories_filename":
        }
        "run_dft":[{
            "image":
            "bohrium":{"scass_type","job_type","platform"},
            "example":[],
            "group_size":,
            "extra_files":[],
            "command":
        }],
        "post_dft":{
            "command":,
            "extra_files":,
            "image":,
            "bohrium":,
            "metrics":{
                "path":,
                "metric_name":
                "value_from_file":,},
            "super_metric":{
                "value_from_file"},
            "upload_datahub":,
            "upload_tracking":
        }
    }
    """
    logs.iprint("read config setting ...")
    config = comm_func.read_config(opts)

    # download examples/predft_examples/rundft_examples/postdft_examples/
    # predft_extrafiles/rundft_extrafiles/postdft_extrafiles
    logs.iprint("read source setting ...")
    datas = comm_class_exampleSource.read_source(opts,work_path,download_path,logs.iprint)
    if datas == None:
        logs.iprint("Error: download examples or rundft_extrafiles or postdft_extrafiles failed!")
        return None  
    
    need_prepare, prepare = comm_class_prepare.construct_input(datas,opts,work_path,download_path,logs)
    #prepare is a dict
    
    need_predft, pre_dft = comm_class_predft.construct_input(datas,opts,logs)
    #pre_dft is a dict

    need_rundft, run_dft = comm_class_rundft.construct_input(datas,opts,logs)
    #run_dft is a list of dict

    need_postdft, post_dft = comm_class_postdft.construct_input(datas,opts,logs)
    #post_dft is a dict
    
    #read metrics setting
    logs.iprint("read metrics setting ...")
    metrics_set = comm_class_metrics.parse_metrics_set(opts)
    if "metrics" in metrics_set:
        post_dft["metrics"] = metrics_set["metrics"]
        if need_prepare:
            # read INPUT parameters setted in prepare, and set it to metrics
            if "mix_input" in prepare and len(prepare["mix_input"]) > 0:
                if "metrics_name" not in post_dft["metrics"]:
                    post_dft["metrics"]["metrics_name"] = []
                post_dft["metrics"]["metrics_name"].append({"INPUT":[]})
                for ikey in prepare["mix_input"]:
                    if "|" in ikey:
                        post_dft["metrics"]["metrics_name"][-1]["INPUT"] += ikey.split("|")
                    else:
                        post_dft["metrics"]["metrics_name"][-1]["INPUT"].append(ikey)
            # do not set path for metrics 
        #elif run_dft[-1].get("example"):
        #    post_dft["metrics"]["path"] = run_dft[-1]["example"]
        #elif post_dft.get("example"):
        #    post_dft["metrics"]["path"] = post_dft.get("example")
        #else:
        #    post_dft["metrics"]["path"] = ["*"]
        
        
 
        need_postdft = True
    if "super_metrics" in metrics_set:
        post_dft["super_metrics"] = metrics_set["super_metrics"]
        need_postdft = True

    #read tracking setting
    if need_postdft and hasattr(opts,"Tracking_metrics"):
        tracking_set = comm_class.TrackingSet.parse_obj(opts)
        if tracking_set:
            config["aim_access_token"] = tracking_set.get("token")
            post_dft["upload_tracking"] = {
                "tags": tracking_set.get("tags"),
                "name": tracking_set.get("name"),
                "experiment": tracking_set.get("experiment")
            }

    allparams = {"config": config,
            "save_path": "results"}
    if need_prepare:
        allparams["prepare"] = prepare
    if need_predft:
        allparams["pre_dft"] = pre_dft
    if need_rundft:
        allparams["run_dft"] = run_dft
    if need_postdft:
        allparams["post_dft"] = post_dft
    logs.iprint("read setting over!\n")
    return allparams
